###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=invalid-name, protected-access
from os import environ
from everysk.config import settings
from everysk.core import firestore, redis as redis_module
from everysk.core.compress import compress
from everysk.core.datetime import Date, DateTime
from everysk.core.exceptions import FieldValueError
from everysk.core.object import BaseDict
from everysk.core.threads import ThreadPool
from everysk.core.unittests import TestCase, mock


def _test_connection(Client: mock.MagicMock):
    doc = firestore.FirestoreClient(project_name='P01', database_name='D01')
    assert doc.connection == Client.return_value

@mock.patch.object(firestore.firestore, 'Client')
class FirestoreClientTestCase(TestCase):

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}

    def test_connections(self, Client: mock.MagicMock):
        doc01 = firestore.FirestoreClient(project_name='P01', database_name='D01')
        self.assertEqual(doc01.project_name, 'P01')
        self.assertEqual(doc01.database_name, 'D01')
        self.assertEqual(doc01.connection, Client.return_value)
        Client.assert_called_once_with(project='P01', database='D01')

        Client.reset_mock()
        doc02 = firestore.FirestoreClient(project_name='P02', database_name='D02')
        self.assertEqual(doc02.project_name, 'P02')
        self.assertEqual(doc02.database_name, 'D02')
        self.assertEqual(doc02.connection, Client.return_value)
        Client.assert_called_once_with(project='P02', database='D02')

    def test_concurrent_connections(self, Client: mock.MagicMock):
        pool = ThreadPool(4)
        for i in range(0, 13): # pylint: disable=unused-variable
            pool.add(
                target=_test_connection,
                kwargs={'Client': Client}
            )
        pool.wait()
        Client.assert_called_once_with(project='P01', database='D01')

    @mock.patch.object(firestore, 'RedisLock')
    def test_connection_lock(self, RedisLock: mock.MagicMock, Client: mock.MagicMock):
        doc01 = firestore.FirestoreClient(project_name='P01', database_name='D01')
        doc01.connection # pylint: disable=pointless-statement
        Client.assert_called_once_with(project='P01', database='D01')
        RedisLock.assert_called_once_with(name='everysk-lib-firestore-lock-connection-P01-D01', timeout=600)
        RedisLock.return_value.acquire.assert_called_once_with(blocking=True)
        RedisLock.return_value.release.assert_called_once_with()

    @mock.patch.object(firestore, 'RedisLock')
    def test_connection_error(self, RedisLock: mock.MagicMock, Client: mock.MagicMock):
        Client.side_effect = AttributeError('Erro')
        doc01 = firestore.FirestoreClient(project_name='P01', database_name='D01')
        with self.assertRaisesRegex(AttributeError, 'Erro'):
            doc01.connection # pylint: disable=pointless-statement
        Client.assert_called_once_with(project='P01', database='D01')
        RedisLock.assert_called_once_with(name='everysk-lib-firestore-lock-connection-P01-D01', timeout=600)
        RedisLock.return_value.acquire.assert_called_once_with(blocking=True)
        RedisLock.return_value.release.assert_called_once_with()

    def test_init_default(self, Client: mock.MagicMock):
        client = firestore.FirestoreClient()
        self.assertEqual(client.database_name, firestore._DEFAULT_DATABASE)
        self.assertEqual(client.project_name, settings.EVERYSK_GOOGLE_CLOUD_PROJECT)
        Client.assert_not_called()

    def test_init_params(self, Client: mock.MagicMock):
        client = firestore.FirestoreClient(project_name='PROJECT', database_name='DATABASE')
        self.assertEqual(client.database_name, 'DATABASE')
        self.assertEqual(client.project_name, 'PROJECT')
        Client.assert_not_called()

    def test_get_collection(self, Client: mock.MagicMock):
        client = firestore.FirestoreClient()
        collection = client.get_collection('my-collection')
        self.assertEqual(collection, Client.return_value.collection.return_value)
        Client.return_value.collection.assert_called_once_with('my-collection')

    def test_environment_values(self, Client: mock.MagicMock):
        environ['EVERYSK_GOOGLE_CLOUD_PROJECT'] = 'PROJECT'
        client = firestore.FirestoreClient()
        self.assertEqual(client.project_name, 'PROJECT')
        Client.assert_not_called()
        del environ['EVERYSK_GOOGLE_CLOUD_PROJECT']


@mock.patch.object(firestore.firestore, 'Client')
class BaseDocumentConfigTestCase(TestCase):

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}

    def test_client(self, Client: mock.MagicMock):
        obj = firestore.BaseDocumentConfig(collection_name='test-collection')
        self.assertIsInstance(obj.client, firestore.FirestoreClient)
        Client.assert_not_called()

    def test_collection(self, Client: mock.MagicMock):
        obj = firestore.BaseDocumentConfig(collection_name='test-collection')
        self.assertEqual(obj.collection, Client.return_value.collection.return_value)
        Client.return_value.collection.assert_called_once_with('test-collection')

    def test_collection_empty(self, Client: mock.MagicMock):
        obj = firestore.BaseDocumentConfig()
        with self.assertRaisesRegex(AttributeError, 'The collection_name is empty.'):
            obj.collection # pylint: disable=pointless-statement
        Client.return_value.collection.assert_not_called()

    def test_excluded_keys(self, Client: mock.MagicMock):
        obj = firestore.BaseDocumentConfig(collection_name='test-collection')
        self.assertListEqual(obj.excluded_keys, [])
        Client.assert_not_called()


class FakeObject:
    """ Used to test pickle and unpickle """
    value: int = 1

    def __eq__(self, __value: object) -> bool:
        return self.__class__ == __value.__class__ and self.value == __value.value


@mock.patch.object(firestore.firestore, 'Client')
class DocumentTestCase(TestCase):

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}
        firestore.Document._config._client = None

    def test_config(self, Client: mock.MagicMock):
        doc = firestore.Document()
        self.assertIsInstance(doc._config, firestore.BaseDocumentConfig)
        Client.assert_not_called()

    @mock.patch.object(firestore.DateTime, 'now', return_value='2022-01-01T10:00:00+00:00')
    def test_init(self, now: mock.MagicMock, Client: mock.MagicMock):
        doc = firestore.Document(key='value', firestore_id='my-id')
        self.assertEqual(doc.firestore_id, 'my-id')
        self.assertEqual(doc.created_at, DateTime.fromisoformat('2022-01-01T10:00:00+00:00'))
        self.assertEqual(doc.key, 'value')
        now.assert_called_once_with()
        Client.assert_not_called()

    def test_parse_in(self, Client: mock.MagicMock):
        doc = firestore.Document(
            dct={'a': 1},
            lst=[1, 2, 3],
            date='2022-01-01',
            datetime='2022-01-01T10:00:00+00:00',
            obj=compress(FakeObject(), serialize='pickle'),
            byte=b'Text',
            base_dict=BaseDict(key='2022-01-01'),
            str_not_date='20240101'
        )
        self.assertDictEqual(doc.dct, {'a': 1})
        self.assertListEqual(doc.lst, [1, 2, 3])
        self.assertEqual(doc.date, Date.fromisoformat('2022-01-01'))
        self.assertEqual(doc.datetime, DateTime.fromisoformat('2022-01-01T10:00:00+00:00'))
        self.assertEqual(doc.obj, FakeObject())
        self.assertEqual(doc.byte, b'Text')
        self.assertDictEqual(doc.base_dict, BaseDict(key=Date.fromisoformat('2022-01-01')))
        self.assertEqual(doc.str_not_date, '20240101')
        Client.assert_not_called()

    def test_parse_in_with_invalid_str_date(self, Client: mock.MagicMock):
        doc = firestore.Document(
            dct={'a': 1},
            lst=[1, 2, 3],
            date='b-a-t',
            datetime='2022-01-01T10:00:00+00:00',
            obj=compress(FakeObject(), serialize='pickle'),
            byte=b'Text',
            base_dict=BaseDict(key='2022-01-01'),
            str_not_date='20240101'
        )
        self.assertDictEqual(doc.dct, {'a': 1})
        self.assertListEqual(doc.lst, [1, 2, 3])
        self.assertEqual(doc.date, 'b-a-t')
        self.assertEqual(doc.datetime, DateTime.fromisoformat('2022-01-01T10:00:00+00:00'))
        self.assertEqual(doc.obj, FakeObject())
        self.assertEqual(doc.byte, b'Text')
        self.assertDictEqual(doc.base_dict, BaseDict(key=Date.fromisoformat('2022-01-01')))
        self.assertEqual(doc.str_not_date, '20240101')
        Client.assert_not_called()

    def test_parser_out(self, Client: mock.MagicMock):
        doc = firestore.Document()
        self.assertDictEqual(doc._parser_out({'a': 1}), {'a': 1})
        self.assertListEqual(doc._parser_out([1, 2, 3]), [1, 2, 3])
        self.assertEqual(doc._parser_out(Date.fromisoformat('2022-01-01')), '2022-01-01')
        self.assertEqual(
            doc._parser_out(DateTime.fromisoformat('2022-01-01T10:00:00+00:00')), '2022-01-01T10:00:00+00:00'
        )
        obj = FakeObject()
        self.assertEqual(doc._parser_out(obj), compress(obj, serialize='pickle'))
        self.assertDictEqual(doc._parser_out(BaseDict(key=Date.fromisoformat('2022-01-01'))), {'key': '2022-01-01'})
        self.assertDictEqual(
            doc._parser_out(BaseDict(key=DateTime.fromisoformat('2022-01-01T10:00:00+00:00'))),
            {'key': '2022-01-01T10:00:00+00:00'}
        )
        Client.assert_not_called()

    def test_loads(self, Client: mock.MagicMock):
        firestore.Document._config.collection_name = 'my-collection'
        Client.return_value.collection.return_value.where.return_value.order_by.return_value.limit.return_value.get.return_value = [
            firestore.Document(firestore_id='id01', type='1', mic='XYSX', created_at='2023-04-03T00:00:00+00:00')
        ]
        doc = firestore.Document.loads(field='firestore_id', condition='==', value='my-id')
        Client.return_value.collection.return_value.where.assert_called_once_with('firestore_id', '==', 'my-id')
        Client.return_value.collection.return_value.where.return_value.order_by.assert_called_once_with('firestore_id')
        self.assertListEqual(
            doc,
            [firestore.Document(**{
                'created_at': DateTime.fromisoformat('2023-04-03T00:00:00+00:00'),
                'firestore_id': 'id01',
                'mic': 'XYSX',
                'type': '1',
                'updated_at': None
            })]
        )
        firestore.Document._config.collection_name = None

    def test_get_firestore_id(self, Client: mock.MagicMock):
        doc = firestore.Document()
        self.assertNotEqual(doc.get_firestore_id(), doc.get_firestore_id())
        doc = firestore.Document(firestore_id='Test')
        self.assertEqual(doc.get_firestore_id(), 'Test')
        Client.assert_not_called()

    def test_load(self, Client: mock.MagicMock):
        firestore.Document._config.collection_name = 'my-collection'
        Client.return_value.collection.return_value.document.return_value.get.return_value.exists = True
        Client.return_value.collection.return_value.document.return_value.get.return_value.to_dict.return_value = {
            'created_at': '2023-04-03T00:00:00+00:00',
            'dct': {'a': 1},
            'lst': [1, 2, 3],
            'date': '2022-01-01',
            'datetime': '2022-01-01T10:00:00+00:00',
            'obj': compress(FakeObject(), serialize='pickle'),
            'byte': b'Text'
        }
        doc = firestore.Document(firestore_id='my-id')
        doc.load()
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        self.assertDictEqual(
            doc,
            {
                'byte': b'Text',
                'created_at': DateTime.fromisoformat('2023-04-03T00:00:00+00:00'),
                'date': Date(2022, 1, 1),
                'datetime': DateTime.fromisoformat('2022-01-01T10:00:00+00:00'),
                'dct': {'a': 1},
                'firestore_id': 'my-id',
                'lst': [1, 2, 3],
                'obj': FakeObject(),
                'updated_at': None
            }
        )
        firestore.Document._config.collection_name = None

    @mock.patch.object(firestore.DateTime, 'now', return_value='2023-10-11T00:00:00+00:00')
    def test_load_no_exists(self, now: mock.MagicMock, Client: mock.MagicMock):
        firestore.Document._config.collection_name = 'my-collection'
        Client.return_value.collection.return_value.document.return_value.get.return_value.exists = False
        Client.return_value.collection.return_value.document.return_value.get.return_value.to_dict.return_value = {
            'created_at': '2023-04-03T00:00:00+00:00',
            'dct': {'a': 1},
            'lst': [1, 2, 3],
            'date': '2022-01-01',
            'datetime': '2022-01-01T10:00:00+00:00',
            'obj': compress(FakeObject(), serialize='pickle'),
            'byte': b'Text'
        }
        doc = firestore.Document(firestore_id='my-id')
        doc.load()
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        self.assertDictEqual(
            doc,
            {
                'created_at': DateTime.fromisoformat('2023-10-11T00:00:00+00:00'),
                'firestore_id': 'my-id',
                'updated_at': None
            }
        )
        firestore.Document._config.collection_name = None
        now.assert_called_once_with()

    @mock.patch.object(firestore.DateTime, 'now', return_value='2023-10-11T00:00:00+00:00')
    def test_save(self, now: mock.MagicMock, Client: mock.MagicMock):
        firestore.Document._config.collection_name = 'my-collection'
        doc = firestore.Document(firestore_id='my-id', var='1', created_at='2023-04-03T00:00:00+00:00')
        doc.save()
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.set.assert_called_once_with(
            document_data={
                'created_at': '2023-04-03T00:00:00+00:00',
                'firestore_id': 'my-id',
                'updated_at': '2023-10-11T00:00:00+00:00',
                'var': '1'
            },
            merge=True,
            timeout=60.0
        )
        now.assert_called_once_with()
        firestore.Document._config.collection_name = None

    def test_to_dict(self, Client: mock.MagicMock):
        obj = FakeObject()
        doc = firestore.Document(
            byte=b'Text',
            created_at=DateTime.fromisoformat('2023-04-03T00:00:00+00:00'),
            date=Date(2022, 1, 1),
            datetime=DateTime.fromisoformat('2022-01-01T10:00:00+00:00'),
            dct={'a': 1},
            firestore_id='my-id',
            lst=[1, 2, 3],
            obj=obj,
            updated_at=None
        )
        self.assertDictEqual(
            doc.to_dict(),
            {
                'byte': b'Text',
                'created_at': '2023-04-03T00:00:00+00:00',
                'date': '2022-01-01',
                'datetime': '2022-01-01T10:00:00+00:00',
                'dct': {'a': 1},
                'firestore_id': 'my-id',
                'lst': [1, 2, 3],
                'obj': compress(obj, serialize='pickle'),
                'updated_at': None
            }
        )
        Client.assert_not_called()

    def test_to_dict_excluded_keys(self, Client: mock.MagicMock):
        doc = firestore.Document(firestore_id='my-id', created_at='2023-04-03T00:00:00+00:00')
        self.assertDictEqual(
            doc.to_dict(),
            {
                'created_at': '2023-04-03T00:00:00+00:00',
                'firestore_id': 'my-id',
                'updated_at': None
            }
        )
        firestore.Document._config.excluded_keys = ['created_at', 'updated_at']
        self.assertDictEqual(doc.to_dict(), {'firestore_id': 'my-id'})
        firestore.Document._config.excluded_keys = []
        Client.assert_not_called()


@mock.patch.object(firestore.firestore, 'Client')
class LoadsPaginatedTestCase(TestCase):
    # pylint: disable=invalid-name

    @classmethod
    def setUpClass(cls) -> None:
        firestore.Document._config.collection_name = 'my-collection'

    @classmethod
    def tearDownClass(cls) -> None:
        firestore.Document._config.collection_name = None

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}
        firestore.Document._config._client = None
        self.response = [
            firestore.Document(firestore_id='id01', type='1', mic='XYSX', created_at='2023-04-03T00:00:00+00:00'),
            firestore.Document(firestore_id='id02', type='2', mic='XZAG', created_at='2023-04-03T00:00:00+00:00'),
            firestore.Document(firestore_id='id03', type='3', mic='XZIM', created_at='2023-04-03T00:00:00+00:00'),
            firestore.Document(firestore_id='id04', type='4', mic='ZARX', created_at='2023-04-03T00:00:00+00:00')
        ]

    def test_loads_paginated_default(self, Client: mock.MagicMock):
        response = self.response.copy()
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.return_value = response
        result = firestore.Document.loads_paginated()
        self.assertListEqual(result, self.response)
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.order_by.assert_called_once_with('firestore_id')
        Client.return_value.collection.return_value.order_by.return_value.limit.assert_called_once_with(500)
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()

    def test_loads_paginated_default_limit(self, Client: mock.MagicMock):
        response = self.response.copy()
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.return_value = response
        result = firestore.Document.loads_paginated(limit=100)
        self.assertListEqual(result, self.response)
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.order_by.assert_called_once_with('firestore_id')
        Client.return_value.collection.return_value.order_by.return_value.limit.assert_called_once_with(100)
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()

    def test_loads_paginated_default_limit_more_query(self, Client: mock.MagicMock):
        response = self.response.copy()
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.return_value = response
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.start_after.return_value.get.return_value = response[0:3]
        result = firestore.Document.loads_paginated(limit=4)
        self.assertListEqual(result, self.response + self.response[0:3])
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.order_by.assert_called_once_with('firestore_id')
        Client.return_value.collection.return_value.order_by.return_value.limit.assert_called_once_with(4)
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.start_after.assert_called_once_with(self.response[-1])
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()

    def test_loads_paginated_default_order_by(self, Client: mock.MagicMock):
        response = self.response.copy()
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.return_value = response
        result = firestore.Document.loads_paginated(order_by='mic', limit=100)
        self.assertListEqual(result, self.response)
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.order_by.assert_called_once_with('mic')
        Client.return_value.collection.return_value.order_by.return_value.limit.assert_called_once_with(100)
        Client.return_value.collection.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()

    def test_loads_paginated_default_order_by_error(self, Client: mock.MagicMock):
        response = self.response.copy()
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.return_value.get.return_value = response
        self.assertRaisesRegex(
            ValueError,
            r"The order_by \(type\) must be in fields\(\['mic'\]\).",
            firestore.Document.loads_paginated,
            fields=['mic'],
            order_by='type',
            limit=100
        )
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_not_called()
        Client.return_value.collection.return_value.select.return_value.order_by.assert_not_called()
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.assert_not_called()
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.return_value.get.assert_not_called()

    def test_loads_paginated_default_fields(self, Client: mock.MagicMock):
        response = self.response.copy()
        new_response = [doc.fromkeys(['firestore_id', 'created_at', 'updated_at', 'mic']) for doc in response]
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.return_value.get.return_value = new_response
        result = firestore.Document.loads_paginated(fields=['mic'], order_by='mic', limit=100)
        self.assertListEqual(result, new_response)
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.select.assert_called_once_with(field_paths=['created_at', 'firestore_id', 'mic', 'updated_at'])
        Client.return_value.collection.return_value.select.return_value.order_by.assert_called_once_with('mic')
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.assert_called_once_with(100)
        Client.return_value.collection.return_value.select.return_value.order_by.return_value.limit.return_value.get.assert_called_once_with()


@mock.patch.object(redis_module.log, 'debug', mock.MagicMock)
class BaseDocumentCachedConfigTestCase(TestCase):

    def test_readonly_attributes(self):
        msg = "The field 'key_prefix' value cannot be changed."
        self.assertRaisesRegex(FieldValueError, msg, firestore.BaseDocumentCachedConfig, key_prefix='BANANA')
        with self.assertRaisesRegex(FieldValueError, msg):
            firestore.BaseDocumentCachedConfig(key_prefix='BANANA')

    def test_cache_property(self):
        obj1 = firestore.BaseDocumentCachedConfig()
        self.assertIsInstance(obj1.cache, firestore.RedisCacheCompressed)
        obj2 = firestore.BaseDocumentCachedConfig()
        self.assertNotEqual(obj1, obj2)

    def test_inheritance(self):
        self.assertEqual(
            firestore.BaseDocumentCachedConfig.__base__,
            firestore.BaseDocumentConfig
        )


@mock.patch.object(firestore.firestore, 'Client')
class DocumentCachedTestCase(TestCase):
    # pylint: disable=invalid-name

    @classmethod
    def setUpClass(cls) -> None:
        firestore.DocumentCached._config.collection_name = 'my-collection'
        cls.redis = redis_module.Redis(host=settings.REDIS_HOST, port=settings.REDIS_PORT)
        cls.key = '86085ac4d100a96d4e1739e87857bd4e266f6a2e4ad62b7ac5bee2aa62f88d2e'

    @classmethod
    def tearDownClass(cls) -> None:
        firestore.DocumentCached._config.collection_name = None

    def setUp(self) -> None:
        firestore.FirestoreClient._connections = {}
        firestore.DocumentCached._config._client = None
        self.redis.delete(self.key)

    def test_init(self, Client: mock.MagicMock):
        doc = firestore.DocumentCached(firestore_id='my-id', created_at='2023-04-03T00:00:00+00:00')
        self.assertDictEqual(
            doc,
            {
                'created_at': DateTime.fromisoformat('2023-04-03T00:00:00+00:00'),
                'firestore_id': 'my-id',
                'redis_key': self.key,
                'updated_at': None
            }
        )
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.get.assert_called_once_with()

    def test_clear_cache_key(self, Client: mock.MagicMock):
        doc = firestore.DocumentCached(firestore_id='my-id')
        self.assertIsNotNone(doc._config.cache.get(doc.get_cache_key()))
        doc.clear_cache_key()
        self.assertIsNone(doc._config.cache.get(doc.get_cache_key()))
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.get.assert_called_once_with()

    def test_get_cache_key(self, Client: mock.MagicMock):
        doc = firestore.DocumentCached(firestore_id='my-id')
        self.assertEqual(doc.get_cache_key(), 'firestore-document-redis-cached-my-collection-my-id')
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.get.assert_called_once_with()

    def test_load_cache(self, Client: mock.MagicMock):
        # The first time that we load, if it is not on Redis we get it from Firestore and stores on Redis
        doc = firestore.DocumentCached(firestore_id='my-id', created_at='2023-11-11T00:00:00+00:00')
        # The second time that we load, it will get the info on Redis and update the object
        doc = firestore.DocumentCached(firestore_id='my-id', created_at='2023-04-03T00:00:00+00:00')
        self.assertDictEqual(
            doc,
            {
                'created_at': DateTime.fromisoformat('2023-11-11T00:00:00+00:00'),
                'firestore_id': 'my-id',
                'redis_key': '86085ac4d100a96d4e1739e87857bd4e266f6a2e4ad62b7ac5bee2aa62f88d2e',
                'updated_at': None
            }
        )
        Client.return_value.collection.assert_called_once_with('my-collection')
        Client.return_value.collection.return_value.document.assert_called_once_with('my-id')
        Client.return_value.collection.return_value.document.return_value.get.assert_called_once_with()

    @mock.patch.object(firestore.DateTime, 'now', return_value='2023-11-11T00:00:00+00:00')
    def test_save(self, now: mock.MagicMock, Client: mock.MagicMock):
        doc = firestore.DocumentCached(firestore_id='my-id', created_at='2023-04-03T00:00:00+00:00')
        self.assertIsNotNone(doc._config.cache.get(doc.get_cache_key()))
        doc.save()
        self.assertIsNone(doc._config.cache.get(doc.get_cache_key()))
        Client.return_value.collection.return_value.document.return_value.set.assert_called_once_with(
            document_data={
                'firestore_id': 'my-id',
                'created_at': '2023-04-03T00:00:00+00:00',
                'updated_at': '2023-11-11T00:00:00+00:00',
                'redis_key': '86085ac4d100a96d4e1739e87857bd4e266f6a2e4ad62b7ac5bee2aa62f88d2e'
            },
            merge=True,
            timeout=60
        )
        now.assert_called_once_with()



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.fields import StrField, IntField, RegexField, ListField

###############################################################################
#   Settings Implementation
###############################################################################

FILE_URL_PATH = StrField(default='/file', readonly=True)
FILE_URL_LENGTH = IntField(default=32, readonly=True)
FILE_HASH_LENGTH = IntField(default=40, readonly=True)

FILE_ID_REGEX = RegexField(default=r'^file_[a-zA-Z0-9]', readonly=True)
FILE_ID_MAX_SIZE = IntField(default=30, readonly=True)
FILE_ID_PREFIX = StrField(default='file_', readonly=True)

FILE_DATA_MAX_SIZE_IN_RAW = IntField(default=int(100 * 1024 * 1024), readonly=True)
FILE_DATA_MAX_SIZE_IN_BASE64 = IntField(default=139810133, readonly=True) # int(FILE_DATA_MAX_SIZE_IN_RAW / 3 * 4) # 13.33MB in BASE64 ~= 10MB in RAW

FILE_CONTENT_TYPES = ListField(
    default=[
        None,
        'application/csv',
        'application/javascript',
        'application/json',
        'application/msword',
        'application/octet-stream',
        'application/pdf',
        'application/vnd.ms-excel',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
        'application/x-zip-compressed',
        'application/xml',
        'application/zip',
        'image/bmp',
        'image/gif',
        'image/jpeg',
        'image/png',
        'image/svg+xml',
        'text/comma-separated-values',
        'text/csv',
        'text/plain',
        'text/x-comma-separated-values',
        'text/xml'
    ],
    readonly=True
)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.core.fields import StrField, IntField, FloatField, RegexField

###############################################################################
#   Settings Implementation
###############################################################################
CUSTOM_INDEX_SYMBOL_REGEX = RegexField(default=r'^CUSTOM:[A-Z0-9_]*$', readonly=True)
CUSTOM_INDEX_SYMBOL_PREFIX = StrField(default='CUSTOM:', readonly=True)

CUSTOM_INDEX_SYMBOL_MIN_SIZE = IntField(default=8, readonly=True) # len(CUSTOM_INDEX_SYMBOL_PREFIX) + 1
CUSTOM_INDEX_SYMBOL_MAX_SIZE = IntField(default=107, readonly=True) # len(CUSTOM_INDEX_SYMBOL_PREFIX) + 100

CUSTOM_INDEX_MIN_DATA_BLOB = IntField(default=1, readonly=True)

CUSTOM_INDEX_BASE_PRICE_MIN_VAL = FloatField(default=-9999999999999.99, readonly=True)
CUSTOM_INDEX_BASE_PRICE_MAX_VAL = FloatField(default=+9999999999999.99, readonly=True)

CUSTOM_INDEX_DATA_TYPE_PRICE = StrField(default='PRICE', readonly=True)
CUSTOM_INDEX_DATA_TYPE_RETURN = StrField(default='RETURN', readonly=True)
CUSTOM_INDEX_DATA_TYPE_RETURN_100 = StrField(default='RETURN_100', readonly=True)

CUSTOM_INDEX_PERIODICITY_MONTHLY = StrField(default='M', readonly=True)
CUSTOM_INDEX_PERIODICITY_DAILY = StrField(default='D', readonly=True)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from datetime import timezone
from zoneinfo import ZoneInfo
from everysk.core.datetime.date import date, Date
from everysk.core.datetime.datetime import datetime, DateTime
from everysk.core.unittests import TestCase, SkipTest


####################################################################################
#                                    Date tests                                    #
####################################################################################
class DateTimeTestCase(TestCase):

    def test_ensure_datetime(self):
        ret = DateTime.ensure(datetime(2023, 1, 1, 13, 59, 59))
        self.assertEqual(ret.__class__, DateTime)
        self.assertEqual(ret, DateTime(2023, 1, 1, 13, 59, 59, tzinfo=ZoneInfo('UTC')))

    def test_ensure_with_zone_info_datetime(self):
        ret = DateTime.ensure(datetime(2023, 1, 1, 13, 57, 58, 59, ZoneInfo('America/New_York')))
        self.assertEqual(ret.__class__, DateTime)
        self.assertEqual(ret, DateTime(2023, 1, 1, 13, 57, 58, 59, ZoneInfo('America/New_York')))

    def test_ensure_datetime_with_nanoseconds(self):
        try:
            from proto.datetime_helpers import DatetimeWithNanoseconds # pylint: disable=import-outside-toplevel
        except ImportError as error:
            raise SkipTest(str(error)) from error

        ret = DateTime.ensure(DatetimeWithNanoseconds(2023, 1, 1, 13, 59, 59, 1234))
        self.assertEqual(ret.__class__, DateTime)
        self.assertEqual(ret, DateTime(2023, 1, 1, 13, 59, 59, 1234))

    def test_ensure_datetime_invalid(self):
        with self.assertRaisesRegex(ValueError, "Invalid instantiation of class 'DateTime' from 'date'"):
            DateTime.ensure(date(2023, 1, 1))

    def test_is_datetime(self):
        self.assertTrue(DateTime.is_datetime(DateTime.now()))
        self.assertTrue(DateTime.is_datetime(DateTime(2023, 1, 1)))
        self.assertTrue(DateTime.is_datetime(DateTime(2023, 1, 1)))
        self.assertFalse(DateTime.is_datetime(Date.today()))
        self.assertFalse(DateTime.is_datetime('date'))
        self.assertFalse(DateTime.is_datetime(2022))

    def test_datetime_instance(self):
        ret = DateTime(year=b'\x07\xe7\t\x0e\x12:\x07\x0b\xca\t', month=None, day=None, hour=None, minute=None, second=None, microsecond=None, tzinfo=None)
        self.assertEqual(ret, DateTime(2023, 9, 14, 18, 58, 7, 772617))

    def test_now_default_time_zone(self):
        now = DateTime.now()
        cmp = datetime.now(timezone.utc)
        self.assertEqual(now.__class__, DateTime)
        self.assertEqual(now.tzinfo, ZoneInfo('UTC'))
        self.assertEqual(now.strftime('%Y%m%d%H%M%S'), cmp.strftime('%Y%m%d%H%M%S'))

    def test_now_custom_time_zone(self):
        now = DateTime.now(ZoneInfo('America/New_York'))
        cmp = datetime.now(ZoneInfo('America/New_York'))
        self.assertEqual(now.__class__, DateTime)
        self.assertEqual(now.tzinfo, ZoneInfo('America/New_York'))
        self.assertEqual(now.strftime('%Y%m%d%H%M%S'), cmp.strftime('%Y%m%d%H%M%S'))

    def test_now_str_time_zone(self):
        now = DateTime.now('UTC')
        cmp = datetime.now(timezone.utc)
        self.assertEqual(now.__class__, DateTime)
        self.assertEqual(now.tzinfo, ZoneInfo('UTC'))
        self.assertEqual(now.strftime('%Y%m%d%H%M%S'), cmp.strftime('%Y%m%d%H%M%S'))

    def test_string_date_to_datetime_with_midday(self):
        result = DateTime(2023, 8, 15).force_time('MIDDAY')
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 8, 15, 12, 0))

    def test_string_date_to_datetime_with_first_minute(self):
        result = DateTime(2023, 8, 15).force_time('FIRST_MINUTE')
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 8, 15, 0, 0))

    def test_string_date_to_datetime_with_now(self):
        date_string = '20230815 ' + datetime.utcnow().strftime('%H:%M:%S.%f')
        expected_result = DateTime.strptime(date_string, '%Y%m%d %H:%M:%S.%f')
        result = DateTime(2023, 8, 15).force_time('NOW').replace(microsecond=expected_result.microsecond)
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, expected_result)

    def test_string_date_to_datetime_with_last_minute(self):
        result = DateTime(2023, 8, 15).force_time('LAST_MINUTE')
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 8, 15, 23, 59, 59, 999999))

    def test_string_date_to_datetime_default_force_time(self):
        result = DateTime(2023, 8, 15).force_time()
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 8, 15, 12, 0))

    def test_string_date_to_datetime_invalid_force_time(self):
        with self.assertRaisesRegex(ValueError, "Invalid force_time. Please choose one of the following: NOW, MIDDAY, FIRST_MINUTE, LAST_MINUTE."):
            DateTime(2023, 8, 15).force_time('INVALID')

    def test_strptime_valid_datetime_string(self):
        result = DateTime.strptime('20230731 15:30:45')
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 7, 31, 15, 30, 45))

    def test_strptime_or_null_valid_datetime_string(self):
        result = DateTime.strptime_or_null('20230731 15:30:45')
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 7, 31, 15, 30, 45))

    def test_strptime_or_null_invalid_datetime_string(self):
        result = DateTime.strptime_or_null('invalid')
        self.assertEqual(result, None)

    def test_string_date_to_datetime_midday(self):
        obj = DateTime.string_date_to_date_time('20230101')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj, DateTime(2023, 1, 1, 12, 0))

    def test_string_date_to_datetime_now(self):
        datetime_now = DateTime.now()
        obj = DateTime.string_date_to_date_time('20230101', 'NOW')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj, DateTime(2023, 1, 1, datetime_now.hour, datetime_now.minute, datetime_now.second))

    def test_string_date_to_datetime_first_minute(self):
        obj = DateTime.string_date_to_date_time('20230101', 'FIRST_MINUTE')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj, DateTime(2023, 1, 1, 0, 0, 0))

    def test_string_date_to_datetime_last_minute(self):
        obj = DateTime.string_date_to_date_time('20230101', 'LAST_MINUTE')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj, DateTime(2023, 1, 1, 23, 59, 59))

    def test_fromisoformat_date(self):
        obj = DateTime.fromisoformat('2023-01-01')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')
        obj = DateTime.fromisoformat('20230101')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')

    def test_fromisoformat_utc(self):
        obj = DateTime.fromisoformat('2023-01-01T00:00:00+00:00')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')
        obj = DateTime.fromisoformat('2023-01-01 00:00:00+00:00')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')

    def test_fromisoformat_brt(self):
        obj = DateTime.fromisoformat('2023-01-01T00:00:00-03:00')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T03:00:00+00:00')
        obj = DateTime.fromisoformat('2023-01-01 00:00:00-03:00')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T03:00:00+00:00')

    def test_fromisoformat_z(self):
        obj = DateTime.fromisoformat('2023-01-01T00:00:00.000000Z')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')
        obj = DateTime.fromisoformat('2023-01-01 00:00:00.000000Z')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')

    def test_fromisoformat_z_with_more(self):
        obj = DateTime.fromisoformat('2023-01-01T00:00:00.0000000Z')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')
        obj = DateTime.fromisoformat('2023-01-01 00:00:00.0000000Z')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')

    def test_fromisoformat_z_with_less(self):
        obj = DateTime.fromisoformat('2023-01-01T00:00:00.00000Z')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')
        obj = DateTime.fromisoformat('2023-01-01 00:00:00.00000Z')
        self.assertEqual(obj.__class__, DateTime)
        self.assertEqual(obj.astimezone(tz='UTC').isoformat(), '2023-01-01T00:00:00+00:00')

    def test_to_isoformat_standard_datetime(self):
        result = DateTime(2023, 7, 31, 12, 34, 56).isoformat()
        self.assertEqual(result, '2023-07-31T12:34:56+00:00')

    def test_to_isoformat_microseconds(self):
        result = DateTime(2023, 1, 1, 0, 0, 0, microsecond=123456).isoformat()
        self.assertEqual(result, '2023-01-01T00:00:00.123456+00:00')

    def test_datetime_to_timestamp_standard_datetime(self):
        result = DateTime(2023, 7, 31, 12, 34, 56).timestamp()
        self.assertEqual(result, 1690806896)

    def test_datetime_to_timestamp_zero_seconds(self):
        result = DateTime(2023, 1, 1, 0, 0, 0).timestamp()
        self.assertEqual(result, 1672531200)

    def test_timestamp_to_datetime_standard(self):
        result = DateTime.fromtimestamp(1699247696)
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 11, 6, 5, 14, 56))

    def test_timestamp_to_datetime_zero_seconds(self):
        result = DateTime.fromtimestamp(1672531200)
        self.assertEqual(result, DateTime(2023, 1, 1, 0, 0))

    def test_datetime_to_datetime_default_time(self):
        result = DateTime.date_to_date_time(Date(2023, 7, 31))
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 7, 31, 12, 0))

    def test_datetime_to_datetime_custom_time(self):
        result = DateTime.date_to_date_time(Date(2023, 1, 1), frc_time='LAST_MINUTE')
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, DateTime(2023, 1, 1, 23, 59, 59, 999999))

    def test_adjust_time_zone_america_new_york(self):
        date_time_obj = DateTime(2023, 7, 31, 12, 34, 56, tzinfo=ZoneInfo('UTC'))
        expected_result = DateTime(2023, 7, 31, 12, 34, 56, tzinfo=ZoneInfo('America/New_York'))
        result = date_time_obj.adjust_time_zone(time_zone='America/New_York')
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, expected_result)

    def test_adjust_time_zone_default_time_zone(self):
        date_time_obj = DateTime(2023, 7, 31, 12, 34, 56, tzinfo=ZoneInfo('UTC'))
        default_result = date_time_obj.adjust_time_zone()
        expected_result = DateTime(2023, 7, 31, 12, 34, 56, tzinfo=ZoneInfo('UTC'))
        self.assertEqual(default_result, expected_result)

    def test_astimezone_america_new_york(self):
        date_time_obj = DateTime(2023, 7, 31, 12, 34, 56, tzinfo=ZoneInfo('UTC'))
        expected_result = DateTime(2023, 7, 31, 8, 34, 56, tzinfo=ZoneInfo('America/New_York'))
        result = date_time_obj.astimezone(tz='America/New_York')
        self.assertEqual(result.__class__, DateTime)
        self.assertEqual(result, expected_result)

    def test_astimezone_default_time_zone(self):
        date_time_obj = DateTime(2023, 7, 31, 12, 34, 56, tzinfo=ZoneInfo('UTC'))
        default_result = date_time_obj.astimezone()
        expected_result = DateTime(2023, 7, 31, 12, 34, 56, tzinfo=ZoneInfo('UTC'))
        self.assertEqual(default_result, expected_result)

    def test_datetime_to_pretty(self):
        result = DateTime(2023, 7, 31, 12, 34, 56).strftime_pretty()
        self.assertEqual(result, 'Jul. 31, 2023, 12:34:56 p.m.')

    def test_datetime_to_pretty_time_only(self):
        result = DateTime(2023, 7, 31, 12, 34, 56).strftime_pretty(just_time=True)
        self.assertEqual(result, '12:34 p.m.')

    def test_datetime_to_pretty_date_only(self):
        result = DateTime(2023, 7, 31, 12, 34, 56).strftime_pretty(just_date=True)
        self.assertEqual(result, 'Jul. 31, 2023')

    def test_datetime_to_pretty_time_pm(self):
        result = DateTime(2023, 7, 31, 12, 34, 56).strftime_pretty(just_time=True)
        self.assertEqual(result, '12:34 p.m.')

    def test_datetime_to_pretty_time_am(self):
        result = DateTime(2023, 1, 1, 0, 0, 0).strftime_pretty(just_time=True)
        self.assertEqual(result, '12:00 a.m.')

    def test_datetime_to_pretty_raise_exception(self):
        with self.assertRaisesRegex(ValueError, 'Both "just_date" and "just_time" flags cannot be true'):
            DateTime(2023, 1, 1, 0, 0, 0).strftime_pretty(just_time=True, just_date=True)

    def test_datetime_to_string_default_format(self):
        result = DateTime(2023, 7, 31, 15, 30, 45).strftime()
        self.assertEqual(result, '20230731 15:30:45')  # Date string in the default format
        self.assertEqual(result, datetime(2023, 7, 31, 15, 30, 45).strftime('%Y%m%d %H:%M:%S'))

    def test_datetime_to_string_custom_format(self):
        custom_format = '%d/%m/%Y %H:%M:%S'
        result = DateTime(2023, 7, 31, 15, 30, 45).strftime(format=custom_format)
        self.assertEqual(result, '31/07/2023 15:30:45')  # Date string using custom format

    def test_date_strftime_or_null(self):
        self.assertEqual(DateTime.strftime_or_null(DateTime(2023, 7, 31, 13, 30, 0)), '20230731 13:30:00')
        self.assertEqual(DateTime.strftime_or_null(DateTime(2023, 7, 31, 13, 30, 0), format='%Y%m%d'), '20230731')
        self.assertEqual(DateTime.strftime_or_null(DateTime(2023, 7, 31, 13, 30, 0), format='%Y-%m-%d'), '2023-07-31')

        self.assertEqual(DateTime.strftime_or_null(datetime(2023, 7, 31, 13, 30, 0)), '20230731 13:30:00')
        self.assertEqual(DateTime.strftime_or_null(datetime(2023, 7, 31, 13, 30, 0), format='%Y%m%d'), '20230731')
        self.assertEqual(DateTime.strftime_or_null(datetime(2023, 7, 31, 13, 30, 0), format='%Y-%m-%d'), '2023-07-31')

        self.assertEqual(DateTime.strftime_or_null(Date(2023, 7, 31)), '20230731 00:00:00')
        self.assertEqual(DateTime.strftime_or_null(Date(2023, 7, 31), format='%Y%m%d'), '20230731')
        self.assertEqual(DateTime.strftime_or_null(Date(2023, 7, 31), format='%Y-%m-%d'), '2023-07-31')

        self.assertEqual(DateTime.strftime_or_null(date(2023, 7, 31)), '20230731 00:00:00')
        self.assertEqual(DateTime.strftime_or_null(date(2023, 7, 31), format='%Y%m%d'), '20230731')
        self.assertEqual(DateTime.strftime_or_null(date(2023, 7, 31), format='%Y-%m-%d'), '2023-07-31')

        self.assertIsNone(DateTime.strftime_or_null('2023-07-31'))
        self.assertIsNone(DateTime.strftime_or_null('2023-07-31', format='%Y%m%d'))
        self.assertIsNone(DateTime.strftime_or_null(None))
        self.assertIsNone(DateTime.strftime_or_null(None, format='%Y-%m-%d'))

    def test_get_hour_standard(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 34, 56).hour, 12)

    def test_get_hour_midnight(self):
        self.assertEqual(DateTime(2023, 1, 1, 0, 0, 0).hour, 0)

    def test_get_minute_standard(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 34, 56).minute, 34)

    def test_get_minute_midnight(self):
        self.assertEqual(DateTime(2023, 1, 1, 0, 0, 0).minute, 0)

    def test_get_second_standard(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 34, 56).second, 56)

    def test_get_second_midnight(self):
        self.assertEqual(DateTime(2023, 1, 1, 0, 0, 0).second, 0)

    def test_is_realtime_portfolio_date_false(self):
        self.assertFalse(DateTime(2023, 1, 1).is_today())

    def test_is_realtime_portfolio_date_true(self):
        self.assertTrue(Date.today().is_today())

    def test_datetime_quarter(self):
        self.assertEqual(DateTime(2023, 1, 1).quarter, 1)
        self.assertEqual(DateTime(2023, 3, 31).quarter, 1)
        self.assertEqual(DateTime(2023, 4, 1).quarter, 2)
        self.assertEqual(DateTime(2023, 6, 30).quarter, 2)
        self.assertEqual(DateTime(2023, 7, 1).quarter, 3)
        self.assertEqual(DateTime(2023, 9, 30).quarter, 3)
        self.assertEqual(DateTime(2023, 10, 1).quarter, 4)
        self.assertEqual(DateTime(2023, 12, 31).quarter, 4)

    def test_datetime_month_name(self):
        self.assertEqual(DateTime(2023, 1, 1).month_name, 'January')
        self.assertEqual(DateTime(2023, 2, 1).month_name, 'February')
        self.assertEqual(DateTime(2023, 3, 1).month_name, 'March')
        self.assertEqual(DateTime(2023, 4, 1).month_name, 'April')
        self.assertEqual(DateTime(2023, 5, 1).month_name, 'May')
        self.assertEqual(DateTime(2023, 6, 1).month_name, 'June')
        self.assertEqual(DateTime(2023, 7, 1).month_name, 'July')
        self.assertEqual(DateTime(2023, 8, 1).month_name, 'August')
        self.assertEqual(DateTime(2023, 9, 1).month_name, 'September')
        self.assertEqual(DateTime(2023, 10, 1).month_name, 'October')
        self.assertEqual(DateTime(2023, 11, 1).month_name, 'November')
        self.assertEqual(DateTime(2023, 12, 1).month_name, 'December')

    def test_datetime_week_of_year(self):
        self.assertEqual(DateTime(2023, 1, 1).week_of_year, 1)
        self.assertEqual(DateTime(2023, 6, 30).week_of_year, 26)
        self.assertEqual(DateTime(2023, 9, 30).week_of_year, 39)
        self.assertEqual(DateTime(2023, 12, 31).week_of_year, 53)

    def test_datetime_week_of_month(self):
        self.assertEqual(DateTime(2023, 1, 1).week_of_month, 1)
        self.assertEqual(DateTime(2023, 6, 8).week_of_month, 2)
        self.assertEqual(DateTime(2023, 9, 15).week_of_month, 3)
        self.assertEqual(DateTime(2023, 12, 31).week_of_month, 5)

    def test_datetime_day_name(self):
        self.assertEqual(DateTime(2023, 1, 1).day_name, 'Sunday')
        self.assertEqual(DateTime(2023, 1, 2).day_name, 'Monday')
        self.assertEqual(DateTime(2023, 1, 3).day_name, 'Tuesday')
        self.assertEqual(DateTime(2023, 1, 4).day_name, 'Wednesday')
        self.assertEqual(DateTime(2023, 1, 5).day_name, 'Thursday')
        self.assertEqual(DateTime(2023, 1, 6).day_name, 'Friday')
        self.assertEqual(DateTime(2023, 1, 7).day_name, 'Saturday')

    def test_datetime_day_of_year(self):
        self.assertEqual(DateTime(2023, 1, 1).day_of_year, 1)
        self.assertEqual(DateTime(2023, 6, 30).day_of_year, 181)
        self.assertEqual(DateTime(2023, 9, 30).day_of_year, 273)
        self.assertEqual(DateTime(2023, 12, 31).day_of_year, 365)

    def test_datetime_toordinal(self):
        self.assertEqual(DateTime(1, 1, 1, 12, 0).toordinal(), 1)
        self.assertEqual(DateTime(1, 12, 31, 12, 0).toordinal(start_date=Date(1, 1, 1)), 365)
        self.assertEqual(DateTime(2134, 12, 28, 12, 0).toordinal(start_date=Date(2020, 1, 1)), 42000)

    def test_datetime_get_last_day_of_week(self):
        ret = DateTime(2023, 1, 1, 12, 0).get_last_day_of_week()
        self.assertEqual(ret, DateTime(2023, 1, 7, 12, 0))
        self.assertEqual(ret.__class__, DateTime)
        self.assertEqual(DateTime(2023, 1, 7, 12, 0).get_last_day_of_week(), DateTime(2023, 1, 7, 12, 0))
        self.assertEqual(DateTime(2023, 1, 7, 12, 0).get_last_day_of_week(bizdays=True), DateTime(2023, 1, 6, 12, 0))
        self.assertEqual(DateTime(2023, 1, 1, 12, 0).get_last_day_of_week(bizdays=True, calendar='ANBIMA'), DateTime(2023, 1, 6, 12, 0))
        self.assertEqual(DateTime(2023, 4, 3, 12, 0).get_last_day_of_week(), DateTime(2023, 4, 8, 12, 0))
        self.assertEqual(DateTime(2023, 4, 3, 12, 0).get_last_day_of_week(bizdays=True), DateTime(2023, 4, 7, 12, 0))
        self.assertEqual(DateTime(2023, 4, 3, 12, 0).get_last_day_of_week(bizdays=True, calendar='ANBIMA'), DateTime(2023, 4, 6, 12, 0))

    def test_datetime_get_last_day_of_month(self):
        self.assertEqual(DateTime(2018, 5, 15, 12, 0).get_last_day_of_month().__class__, DateTime)
        self.assertEqual(DateTime(2018, 5, 15, 12, 0).get_last_day_of_month(), DateTime(2018, 5, 31, 12, 0))
        self.assertEqual(DateTime(2018, 5, 15, 12, 0).get_last_day_of_month(bizdays=True), DateTime(2018, 5, 31, 12, 0))
        self.assertEqual(DateTime(2018, 5, 15, 12, 0).get_last_day_of_month(bizdays=True, calendar='ANBIMA'), DateTime(2018, 5, 30, 12, 0))

    def test_datetime_get_last_day_of_quarter(self):
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_last_day_of_quarter().__class__, DateTime)
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_last_day_of_quarter(), DateTime(2023, 3, 31, 12, 0))
        self.assertEqual(DateTime(2023, 4, 15, 12, 0).get_last_day_of_quarter(), DateTime(2023, 6, 30, 12, 0))
        self.assertEqual(DateTime(2023, 7, 15, 12, 0).get_last_day_of_quarter(bizdays=True), DateTime(2023, 9, 29, 12, 0))
        self.assertEqual(DateTime(2023, 11, 15, 12, 0).get_last_day_of_quarter(bizdays=True, calendar='ANBIMA'), DateTime(2023, 12, 29, 12, 0))

    def test_datetime_get_last_day_of_year(self):
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_last_day_of_year().__class__, DateTime)
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_last_day_of_year(), DateTime(2023, 12, 31, 12, 0))
        self.assertEqual(DateTime(2023, 7, 15, 12, 0).get_last_day_of_year(bizdays=True), DateTime(2023, 12, 29, 12, 0))
        self.assertEqual(DateTime(2023, 11, 15, 12, 0).get_last_day_of_year(bizdays=True, calendar='ANBIMA'), DateTime(2023, 12, 29, 12, 0))

    def test_datetime_is_last_day_of_week(self):
        self.assertTrue(DateTime(2023, 4, 8, 12, 0).is_last_day_of_week())
        self.assertFalse(DateTime(2023, 4, 7, 12, 0).is_last_day_of_week())
        self.assertFalse(DateTime(2023, 4, 6, 12, 0).is_last_day_of_week())
        self.assertFalse(DateTime(2023, 4, 8, 12, 0).is_last_day_of_week(bizdays=True))
        self.assertTrue(DateTime(2023, 4, 7, 12, 0).is_last_day_of_week(bizdays=True))
        self.assertTrue(DateTime(2023, 4, 6, 12, 0).is_last_day_of_week(bizdays=True, calendar='ANBIMA'))

    def test_datetime_is_last_day_of_month(self):
        self.assertTrue(DateTime(2023, 1, 31, 12, 0).is_last_day_of_month())
        self.assertFalse(DateTime(2023, 2, 27, 12, 0).is_last_day_of_month())
        self.assertTrue(DateTime(2023, 12, 31, 12, 0).is_last_day_of_month())
        self.assertTrue(DateTime(2023, 4, 28, 12, 0).is_last_day_of_month(bizdays=True))
        self.assertFalse(DateTime(2023, 4, 30, 12, 0).is_last_day_of_month(bizdays=True))
        self.assertFalse(DateTime(2023, 12, 31, 12, 0).is_last_day_of_month(bizdays=True, calendar='ANBIMA'))

    def test_datetime_is_last_day_of_quarter(self):
        self.assertFalse(DateTime(2023, 1, 1, 12, 0).is_last_day_of_quarter())
        self.assertTrue(DateTime(2023, 3, 31, 12, 0).is_last_day_of_quarter())
        self.assertFalse(DateTime(2023, 4, 1, 12, 0).is_last_day_of_quarter())
        self.assertTrue(DateTime(2023, 6, 30, 12, 0).is_last_day_of_quarter())
        self.assertFalse(DateTime(2023, 7, 1, 12, 0).is_last_day_of_quarter())
        self.assertTrue(DateTime(2023, 9, 30, 12, 0).is_last_day_of_quarter())
        self.assertFalse(DateTime(2023, 10, 1, 12, 0).is_last_day_of_quarter())
        self.assertTrue(DateTime(2023, 12, 31, 12, 0).is_last_day_of_quarter())
        self.assertTrue(DateTime(2023, 12, 29, 12, 0).is_last_day_of_quarter(bizdays=True))
        self.assertTrue(DateTime(2023, 12, 29, 12, 0).is_last_day_of_quarter(bizdays=True, calendar='ANBIMA'))
        self.assertFalse(DateTime(2023, 12, 31, 12, 0).is_last_day_of_quarter(bizdays=True))
        self.assertFalse(DateTime(2023, 12, 31, 12, 0).is_last_day_of_quarter(bizdays=True, calendar='ANBIMA'))

    def test_datetime_is_last_day_of_year(self):
        self.assertFalse(DateTime(2023, 1, 1, 12, 0).is_last_day_of_year())
        self.assertFalse(DateTime(2023, 1, 31, 12, 0).is_last_day_of_year())
        self.assertTrue(DateTime(2023, 12, 31, 12, 0).is_last_day_of_year())
        self.assertTrue(DateTime(2023, 12, 29, 12, 0).is_last_day_of_year(bizdays=True))
        self.assertTrue(DateTime(2023, 12, 29, 12, 0).is_last_day_of_year(bizdays=True, calendar='ANBIMA'))
        self.assertFalse(DateTime(2023, 12, 30, 12, 0).is_last_day_of_year(bizdays=True))
        self.assertFalse(DateTime(2023, 12, 31, 12, 0).is_last_day_of_year(bizdays=True, calendar='ANBIMA'))

    def test_datetime_get_first_day_of_week(self):
        self.assertEqual(DateTime(2023, 1, 1, 12, 0).get_first_day_of_week().__class__, DateTime)
        self.assertEqual(DateTime(2023, 1, 1, 12, 0).get_first_day_of_week(), DateTime(2023, 1, 1, 12, 0))
        self.assertEqual(DateTime(2023, 1, 5, 12, 0).get_first_day_of_week(), DateTime(2023, 1, 1, 12, 0))
        self.assertEqual(DateTime(2023, 1, 7, 12, 0).get_first_day_of_week(bizdays=True), DateTime(2023, 1, 2, 12, 0))
        self.assertEqual(DateTime(2023, 1, 1, 12, 0).get_first_day_of_week(bizdays=True, calendar='ANBIMA'), DateTime(2023, 1, 2, 12, 0))
        self.assertEqual(DateTime(2023, 4, 3, 12, 0).get_first_day_of_week(), DateTime(2023, 4, 2, 12, 0))
        self.assertEqual(DateTime(2023, 4, 3, 12, 0).get_first_day_of_week(bizdays=True), DateTime(2023, 4, 3, 12, 0))
        self.assertEqual(DateTime(2023, 4, 3, 12, 0).get_first_day_of_week(bizdays=True, calendar='ANBIMA'), DateTime(2023, 4, 3, 12, 0))
        self.assertEqual(DateTime(2023, 5, 4, 12, 0).get_first_day_of_week(bizdays=True, calendar='ANBIMA'), DateTime(2023, 5, 2, 12, 0))

    def test_datetime_get_first_day_of_month(self):
        self.assertEqual(DateTime(2023, 5, 15, 12, 0).get_first_day_of_month().__class__, DateTime)
        self.assertEqual(DateTime(2023, 5, 15, 12, 0).get_first_day_of_month(), DateTime(2023, 5, 1, 12, 0))
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_first_day_of_month(bizdays=True), DateTime(2023, 1, 2, 12, 0))
        self.assertEqual(DateTime(2023, 5, 15, 12, 0).get_first_day_of_month(bizdays=True, calendar='ANBIMA'), DateTime(2023, 5, 2, 12, 0))

    def test_datetime_get_first_day_of_quarter(self):
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_first_day_of_quarter().__class__, DateTime)
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_first_day_of_quarter(), DateTime(2023, 1, 1, 12, 0))
        self.assertEqual(DateTime(2023, 4, 15, 12, 0).get_first_day_of_quarter(), DateTime(2023, 4, 1, 12, 0))
        self.assertEqual(DateTime(2023, 4, 15, 12, 0).get_first_day_of_quarter(bizdays=True), DateTime(2023, 4, 3, 12, 0))
        self.assertEqual(DateTime(2023, 11, 15, 12, 0).get_first_day_of_quarter(bizdays=True, calendar='ANBIMA'), DateTime(2023, 10, 2, 12, 0))

    def test_datetime_get_first_day_of_year(self):
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_first_day_of_year().__class__, DateTime)
        self.assertEqual(DateTime(2023, 1, 15, 12, 0).get_first_day_of_year(), DateTime(2023, 1, 1, 12, 0))
        self.assertEqual(DateTime(2023, 7, 15, 12, 0).get_first_day_of_year(bizdays=True), DateTime(2023, 1, 2, 12, 0))
        self.assertEqual(DateTime(2018, 11, 15, 12, 0).get_first_day_of_year(bizdays=True, calendar='ANBIMA'), DateTime(2018, 1, 2, 12, 0))

    def test_datetime_is_first_day_of_week(self):
        self.assertTrue(DateTime(2023, 1, 1, 12, 0).is_first_day_of_week())
        self.assertTrue(DateTime(2023, 4, 2, 12, 0).is_first_day_of_week())
        self.assertFalse(DateTime(2023, 1, 1, 12, 0).is_first_day_of_week(bizdays=True))
        self.assertTrue(DateTime(2018, 1, 2, 12, 0).is_first_day_of_week(bizdays=True, calendar='ANBIMA'))

    def test_datetime_is_first_day_of_month(self):
        self.assertTrue(DateTime(2023, 1, 1, 12, 0).is_first_day_of_month())
        self.assertTrue(DateTime(2023, 12, 1, 12, 0).is_first_day_of_month())
        self.assertFalse(DateTime(2023, 12, 31, 12, 0).is_first_day_of_month())
        self.assertTrue(DateTime(2023, 1, 2, 12, 0).is_first_day_of_month(bizdays=True))
        self.assertFalse(DateTime(2023, 1, 1, 12, 0).is_first_day_of_month(bizdays=True))
        self.assertTrue(DateTime(2018, 1, 2, 12, 0).is_first_day_of_month(bizdays=True, calendar='ANBIMA'))
        self.assertFalse(DateTime(2018, 1, 1, 12, 0).is_first_day_of_month(bizdays=True, calendar='ANBIMA'))

    def test_datetime_is_first_day_of_quarter(self):
        self.assertTrue(DateTime(2023, 1, 1, 12, 0).is_first_day_of_quarter())
        self.assertFalse(DateTime(2023, 3, 31, 12, 0).is_first_day_of_quarter())
        self.assertTrue(DateTime(2023, 4, 1, 12, 0).is_first_day_of_quarter())
        self.assertFalse(DateTime(2023, 6, 30, 12, 0).is_first_day_of_quarter())
        self.assertTrue(DateTime(2023, 7, 1, 12, 0).is_first_day_of_quarter())
        self.assertFalse(DateTime(2023, 9, 30, 12, 0).is_first_day_of_quarter())
        self.assertTrue(DateTime(2023, 10, 1, 12, 0).is_first_day_of_quarter())
        self.assertFalse(DateTime(2023, 12, 31, 12, 0).is_first_day_of_quarter())
        self.assertTrue(DateTime(2023, 1, 2, 12, 0).is_first_day_of_quarter(bizdays=True))
        self.assertTrue(DateTime(2018, 7, 2, 12, 0).is_first_day_of_quarter(bizdays=True))
        self.assertTrue(DateTime(2018, 1, 2, 12, 0).is_first_day_of_quarter(bizdays=True, calendar='ANBIMA'))

    def test_datetime_is_first_day_of_year(self):
        self.assertTrue(DateTime(2023, 1, 1, 12, 0).is_first_day_of_year())
        self.assertTrue(DateTime(2023, 1, 2, 12, 0).is_first_day_of_year(bizdays=True))
        self.assertTrue(DateTime(2018, 1, 2, 12, 0).is_first_day_of_year(bizdays=True, calendar='ANBIMA'))
        self.assertFalse(DateTime(2023, 12, 30, 12, 0).is_first_day_of_year(bizdays=True))
        self.assertFalse(DateTime(2023, 12, 31, 12, 0).is_first_day_of_year(bizdays=True, calendar='ANBIMA'))

    def test_datetime_delta(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=5, periodicity='D').__class__, DateTime)
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=5, periodicity='D'), DateTime(2023, 8, 5, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=2, periodicity='W'), DateTime(2023, 8, 14, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=1, periodicity='M'), DateTime(2023, 8, 31, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=2, periodicity='Y'), DateTime(2025, 7, 31, 12, 0))
        self.assertEqual(DateTime(2023, 4, 30, 12, 0).delta(period=1, periodicity='B', calendar='ANBIMA'), DateTime(2023, 5, 2, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=5.5, periodicity='D'), DateTime(2023, 8, 6, 0, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=2.5, periodicity='W'), DateTime(2023, 8, 18, 0, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=1.5, periodicity='M'), DateTime(2023, 8, 31, 12, 0)) # ignoring float
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=2.5, periodicity='Y'), DateTime(2025, 7, 31, 12, 0)) # ignoring float
        self.assertEqual(DateTime(2023, 4, 30, 12, 0).delta(period=1.5, periodicity='B', calendar='ANBIMA'), DateTime(2023, 5, 3, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=-5.5, periodicity='D'), DateTime(2023, 7, 26, 0, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=-2.5, periodicity='W'), DateTime(2023, 7, 14, 0, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=-1.5, periodicity='M'), DateTime(2023, 6, 30, 12, 0)) # ignoring float
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=-2.5, periodicity='Y'), DateTime(2021, 7, 31, 12, 0)) # ignoring float
        self.assertEqual(DateTime(2023, 4, 30, 12, 0).delta(period=-1.5, periodicity='B', calendar='ANBIMA'), DateTime(2023, 4, 27, 12, 0))

        with self.assertRaisesRegex(ValueError, "Invalid periodicity. Please choose one of the following: D, B, W, M, Y."):
            self.assertEqual(DateTime(2023, 7, 31, 12, 0).delta(period=5, periodicity='INVALID'), DateTime(2023, 8, 5, 12, 0))

        with self.assertRaisesRegex(ValueError, "Value out of range"):
            self.assertEqual(Date(2023, 7, 31).days_delta(50001), Date(2023, 8, 5))

    def test_datetime_days_delta(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).days_delta(5).__class__, DateTime)
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).days_delta(5), DateTime(2023, 8, 5, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).days_delta(-5), DateTime(2023, 7, 26, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).days_delta(5.5), DateTime(2023, 8, 6, 0, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).days_delta(-5.5), DateTime(2023, 7, 26, 0, 0))

        with self.assertRaisesRegex(ValueError, "Value out of range"):
            self.assertEqual(DateTime(2023, 7, 31, 12, 0).days_delta(50001), DateTime(2023, 8, 5, 12, 0))

    def test_datetime_bizdays_delta(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).bizdays_delta(5).__class__, DateTime)
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).bizdays_delta(5), DateTime(2023, 8, 7, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).bizdays_delta(-5), DateTime(2023, 7, 24, 12, 0))
        self.assertEqual(DateTime(2023, 4, 30, 12, 0).bizdays_delta(1, calendar='ANBIMA'), DateTime(2023, 5, 2, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).bizdays_delta(5.5), DateTime(2023, 8, 8, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).bizdays_delta(-5.5), DateTime(2023, 7, 21, 12, 0))
        self.assertEqual(DateTime(2023, 4, 30, 12, 0).bizdays_delta(1.5, calendar='ANBIMA'), DateTime(2023, 5, 3, 12, 0))

        with self.assertRaisesRegex(ValueError, "Value out of range"):
            self.assertEqual(DateTime(2023, 7, 31, 12, 0).bizdays_delta(50001), DateTime(2023, 8, 5, 12, 0))

    def test_datetime_weeks_delta(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).weeks_delta(5).__class__, DateTime)
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).weeks_delta(5), DateTime(2023, 9, 4, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).weeks_delta(-5), DateTime(2023, 6, 26, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).weeks_delta(5.5), DateTime(2023, 9, 8, 0, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).weeks_delta(-5.5), DateTime(2023, 6, 23, 0, 0))

    def test_datetime_months_delta(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).months_delta(5).__class__, DateTime)
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).months_delta(5), DateTime(2023, 12, 31, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).months_delta(-5), DateTime(2023, 2, 28, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).months_delta(5.5), DateTime(2023, 12, 31, 12, 0)) # ignoring float
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).months_delta(-5.5), DateTime(2023, 2, 28, 12, 0)) # ignoring float

    def test_datetime_months_delta_with_leap_year(self):
        self.assertEqual(DateTime(2020, 3, 31, 12, 0).months_delta(1), DateTime(2020, 4, 30, 12, 0))
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).months_delta(1), DateTime(2020, 3, 29, 12, 0))
        self.assertEqual(DateTime(2020, 3, 31, 12, 0).months_delta(-1), DateTime(2020, 2, 29, 12, 0))
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).months_delta(12), DateTime(2021, 2, 28, 12, 0))
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).months_delta(-12), DateTime(2019, 2, 28, 12, 0))
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).months_delta(12.5), DateTime(2021, 2, 28, 12, 0)) # ignoring float
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).months_delta(-12.5), DateTime(2019, 2, 28, 12, 0)) # ignoring float

    def test_datetime_years_delta(self):
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).years_delta(5).__class__, DateTime)
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).years_delta(5), DateTime(2028, 7, 31, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).years_delta(-5), DateTime(2018, 7, 31, 12, 0))
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).years_delta(5.5), DateTime(2028, 7, 31, 12, 0)) # ignoring float
        self.assertEqual(DateTime(2023, 7, 31, 12, 0).years_delta(-5.5), DateTime(2018, 7, 31, 12, 0)) # ignoring float

    def test_datetime_years_delta_with_leap_year(self):
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).years_delta(1), DateTime(2021, 2, 28, 12, 0))
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).years_delta(-1), DateTime(2019, 2, 28, 12, 0))
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).years_delta(4), DateTime(2024, 2, 29, 12, 0))
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).years_delta(-4), DateTime(2016, 2, 29, 12, 0))
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).years_delta(4.5), DateTime(2024, 2, 29, 12, 0)) # ignoring float
        self.assertEqual(DateTime(2020, 2, 29, 12, 0).years_delta(-4.5), DateTime(2016, 2, 29, 12, 0)) # ignoring float

    def test_datetime_diff(self):
        self.assertEqual(DateTime.diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 7, 31, 12, 0), periodicity='D'), 0)
        self.assertEqual(DateTime.diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 8, 5, 12, 0), periodicity='D'), 5)
        self.assertEqual(DateTime.diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 8, 14, 12, 0), periodicity='W'), 2)
        self.assertEqual(DateTime.diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 8, 31, 12, 0), periodicity='M'), 1)
        self.assertEqual(DateTime.diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2025, 7, 31, 12, 0), periodicity='Y'), 2)
        self.assertEqual(DateTime.diff(start_date=DateTime(2023, 4, 30, 12, 0), end_date=DateTime(2023, 5, 2, 12, 0), periodicity='B', calendar='ANBIMA'), 1)
        self.assertEqual(DateTime.diff(start_date=datetime(2023, 4, 30, 12, 0), end_date=datetime(2023, 5, 2, 12, 0), periodicity='B', calendar='ANBIMA'), 1)

        with self.assertRaisesRegex(ValueError, "Invalid periodicity. Please choose one of the following: D, B, W, M, Y."):
            self.assertEqual(DateTime.diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 8, 5, 12, 0), periodicity='INVALID'), 5)

    def test_datetime_days_diff(self):
        self.assertEqual(DateTime.days_diff(start_date=DateTime(2023, 8, 5, 12, 0), end_date=DateTime(2023, 7, 31, 12, 0)), -5)
        self.assertEqual(DateTime.days_diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 8, 5, 12, 0)), 5)
        self.assertEqual(DateTime.days_diff(start_date=datetime(2023, 7, 31, 12, 0), end_date=datetime(2023, 8, 5, 12, 0)), 5)
        self.assertEqual(DateTime.days_diff(start_date=DateTime(2024, 2, 19, 12, 0), end_date=DateTime(2024, 2, 20, 12, 0)), 1)
        self.assertEqual(DateTime.days_diff(start_date=DateTime(2023, 5, 2, 12, 0), end_date=DateTime(2023, 4, 30, 12, 0)), -2)

    def test_datetime_bizdays_diff(self):
        self.assertEqual(DateTime.bizdays_diff(start_date=DateTime(2023, 8, 7, 12, 0), end_date=DateTime(2023, 7, 31, 12, 0)), -5)
        self.assertEqual(DateTime.bizdays_diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 8, 7, 12, 0)), 5)
        self.assertEqual(DateTime.bizdays_diff(start_date=DateTime(2023, 5, 2, 12, 0), end_date=DateTime(2023, 4, 30, 12, 0), calendar='ANBIMA'), -1)
        self.assertEqual(DateTime.bizdays_diff(start_date=datetime(2023, 5, 2, 12, 0), end_date=datetime(2023, 4, 30, 12, 0), calendar='ANBIMA'), -1)
        self.assertEqual(DateTime.bizdays_diff(start_date=DateTime(2024, 2, 19, 12, 0), end_date=DateTime(2024, 2, 20, 12, 0), calendar='ANBIMA'), 1)
        self.assertEqual(DateTime.bizdays_diff(start_date=DateTime(2023, 5, 2, 12, 0), end_date=DateTime(2023, 4, 30, 12, 0), calendar='ANBIMA'), -1)

    def test_datetime_weeks_diff(self):
        self.assertEqual(DateTime.weeks_diff(start_date=DateTime(2023, 9, 4, 12, 0), end_date=DateTime(2023, 7, 31, 12, 0)), -5)
        self.assertEqual(DateTime.weeks_diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 9, 4, 12, 0)), 5)
        self.assertEqual(DateTime.weeks_diff(start_date=datetime(2023, 7, 31, 12, 0), end_date=datetime(2023, 9, 4, 12, 0)), 5)

    def test_datetime_months_diff(self):
        self.assertEqual(DateTime.months_diff(start_date=DateTime(2023, 12, 31, 12, 0), end_date=DateTime(2023, 7, 31, 12, 0)), -5)
        self.assertEqual(DateTime.months_diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 12, 31, 12, 0)), 5)
        self.assertEqual(DateTime.months_diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2023, 12, 29, 12, 0)), 4)
        self.assertEqual(DateTime.months_diff(start_date=datetime(2023, 7, 31, 12, 0), end_date=datetime(2023, 12, 29, 12, 0)), 4)

    def test_datetime_years_diff(self):
        self.assertEqual(DateTime.years_diff(start_date=DateTime(2028, 7, 31, 12, 0), end_date=DateTime(2023, 7, 31, 12, 0)), -5)
        self.assertEqual(DateTime.years_diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2028, 7, 31, 12, 0)), 5)
        self.assertEqual(DateTime.years_diff(start_date=DateTime(2023, 7, 31, 12, 0), end_date=DateTime(2028, 7, 29, 12, 0)), 4)
        self.assertEqual(DateTime.years_diff(start_date=datetime(2023, 7, 31, 12, 0), end_date=datetime(2028, 7, 29, 12, 0)), 4)

    def test_datetime_range(self):
        self.assertEqual(DateTime.range(DateTime(2023, 1, 1), DateTime(2023, 1, 1)), [])
        self.assertListEqual(
            DateTime.range(DateTime(2021, 12, 23, 12, 0), DateTime(2022, 1, 2, 12, 0)),
            [
                DateTime(2021, 12, 23, 12, 0),
                DateTime(2021, 12, 24, 12, 0),
                DateTime(2021, 12, 25, 12, 0),
                DateTime(2021, 12, 26, 12, 0),
                DateTime(2021, 12, 27, 12, 0),
                DateTime(2021, 12, 28, 12, 0),
                DateTime(2021, 12, 29, 12, 0),
                DateTime(2021, 12, 30, 12, 0),
                DateTime(2021, 12, 31, 12, 0),
                DateTime(2022, 1, 1, 12, 0)
            ]
        )

        self.assertEqual(
            DateTime.range(datetime(2021, 12, 23, 12, 0), datetime(2022, 1, 2, 12, 0)),
            [
                DateTime(2021, 12, 23, 12, 0),
                DateTime(2021, 12, 24, 12, 0),
                DateTime(2021, 12, 25, 12, 0),
                DateTime(2021, 12, 26, 12, 0),
                DateTime(2021, 12, 27, 12, 0),
                DateTime(2021, 12, 28, 12, 0),
                DateTime(2021, 12, 29, 12, 0),
                DateTime(2021, 12, 30, 12, 0),
                DateTime(2021, 12, 31, 12, 0),
                DateTime(2022, 1, 1, 12, 0)
            ]
        )

        self.assertEqual(
            DateTime.range(DateTime(2023, 4, 18, 12, 0), DateTime(2023, 5, 3, 12, 0), periodicity='B', calendar='ANBIMA'),
            [
                DateTime(2023, 4, 18, 12, 0),
                DateTime(2023, 4, 19, 12, 0),
                DateTime(2023, 4, 20, 12, 0),
                DateTime(2023, 4, 24, 12, 0),
                DateTime(2023, 4, 25, 12, 0),
                DateTime(2023, 4, 26, 12, 0),
                DateTime(2023, 4, 27, 12, 0),
                DateTime(2023, 4, 28, 12, 0),
                DateTime(2023, 5, 2, 12, 0)]
        )

    def test_datetime_range_invalid_periodicity(self):
        with self.assertRaisesRegex(ValueError, 'Invalid periodicity. Please choose one of the following: D, B.'):
            DateTime.range(DateTime(2021, 12, 23, 12, 0), DateTime(2022, 1, 2, 12, 0), periodicity='invalid')

    def test_datetime_range_out_of_range_date(self):
        with self.assertRaisesRegex(ValueError, 'Value out of range'):
            DateTime.range(DateTime(2000, 1, 1, 12, 0), DateTime(2300, 1, 1, 12, 0))

    def test_datetime_days_range(self):
        self.assertEqual(DateTime.days_range(DateTime(2023, 1, 1, 12, 0), DateTime(2023, 1, 1, 12, 0)), [])
        self.assertListEqual(
            DateTime.days_range(DateTime(2021, 12, 23, 12, 0), DateTime(2022, 1, 2, 12, 0)),
            [
                DateTime(2021, 12, 23, 12, 0),
                DateTime(2021, 12, 24, 12, 0),
                DateTime(2021, 12, 25, 12, 0),
                DateTime(2021, 12, 26, 12, 0),
                DateTime(2021, 12, 27, 12, 0),
                DateTime(2021, 12, 28, 12, 0),
                DateTime(2021, 12, 29, 12, 0),
                DateTime(2021, 12, 30, 12, 0),
                DateTime(2021, 12, 31, 12, 0),
                DateTime(2022, 1, 1, 12, 0)
            ]
        )

        self.assertListEqual(
            DateTime.days_range(datetime(2021, 12, 23, 12, 0), datetime(2022, 1, 2, 12, 0)),
            [
                DateTime(2021, 12, 23, 12, 0),
                DateTime(2021, 12, 24, 12, 0),
                DateTime(2021, 12, 25, 12, 0),
                DateTime(2021, 12, 26, 12, 0),
                DateTime(2021, 12, 27, 12, 0),
                DateTime(2021, 12, 28, 12, 0),
                DateTime(2021, 12, 29, 12, 0),
                DateTime(2021, 12, 30, 12, 0),
                DateTime(2021, 12, 31, 12, 0),
                DateTime(2022, 1, 1, 12, 0)
            ]
        )

    def test_datetime_bizdays_range_with_calendar(self):
        self.assertListEqual(
            DateTime.bizdays_range(DateTime(2023, 4, 18, 12, 0), DateTime(2023, 5, 3, 12, 0), calendar='ANBIMA'),
            [
                DateTime(2023, 4, 18, 12, 0),
                DateTime(2023, 4, 19, 12, 0),
                DateTime(2023, 4, 20, 12, 0),
                DateTime(2023, 4, 24, 12, 0),
                DateTime(2023, 4, 25, 12, 0),
                DateTime(2023, 4, 26, 12, 0),
                DateTime(2023, 4, 27, 12, 0),
                DateTime(2023, 4, 28, 12, 0),
                DateTime(2023, 5, 2, 12, 0)]
        )

        self.assertListEqual(
            DateTime.bizdays_range(datetime(2023, 4, 18, 12, 0, tzinfo=ZoneInfo('UTC')), datetime(2023, 5, 3, 12, 0, tzinfo=ZoneInfo('UTC')), calendar='ANBIMA'),
            [
                DateTime(2023, 4, 18, 12, 0),
                DateTime(2023, 4, 19, 12, 0),
                DateTime(2023, 4, 20, 12, 0),
                DateTime(2023, 4, 24, 12, 0),
                DateTime(2023, 4, 25, 12, 0),
                DateTime(2023, 4, 26, 12, 0),
                DateTime(2023, 4, 27, 12, 0),
                DateTime(2023, 4, 28, 12, 0),
                DateTime(2023, 5, 2, 12, 0)]
        )

    def test_datetime_bizdays_range_without_calendar(self):
        self.assertEqual(
            DateTime.bizdays_range(DateTime(2023, 4, 18, 12, 0), DateTime(2023, 5, 3, 12, 0)),
            [
                DateTime(2023, 4, 18, 12, 0),
                DateTime(2023, 4, 19, 12, 0),
                DateTime(2023, 4, 20, 12, 0),
                DateTime(2023, 4, 21, 12, 0),
                DateTime(2023, 4, 24, 12, 0),
                DateTime(2023, 4, 25, 12, 0),
                DateTime(2023, 4, 26, 12, 0),
                DateTime(2023, 4, 27, 12, 0),
                DateTime(2023, 4, 28, 12, 0),
                DateTime(2023, 5, 1, 12, 0),
                DateTime(2023, 5, 2, 12, 0)]
        )



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import Date
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock

from everysk.sdk.engines.market_data import MarketData


###############################################################################
#   Market Data Test Case Implementation
###############################################################################
class MarketDataTestCase(TestCase):

    def setUp(self) -> None:
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()
        return super().setUp()

    ###############################################################################
    #  Get Security Test Case Implementation
    ###############################################################################
    def test_get_security_is_called(self):
        market_data = MarketData(date=Date(2023, 7, 31))
        expected_data = compress({'class_name': 'MarketData', 'method_name': 'get_security', 'self_obj': market_data.to_dict(add_class_path=True), 'params': {'date': '123', 'ticker_list': ['test'], 'ticker_type': ['choice1', 'choice2'], 'projection': 'test_projection', 'nearest': True}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            market_data.get_security(date='123', ticker_list=['test'], ticker_type=['choice1', 'choice2'], projection='test_projection', nearest=True)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    ###############################################################################
    #  Get Historical Test Case Implementation
    ###############################################################################
    def test_get_historical_is_called(self):
        market_data = MarketData(date=Date(2023, 7, 31))
        expected_data = compress({'class_name': 'MarketData', 'method_name': 'get_historical', 'self_obj': market_data.to_dict(add_class_path=True), 'params': {'date': '2023-08-14', 'start_date': '2022-05-14', 'end_date': '2024-07-17', 'ticker_list': ['test'], 'ticker_type': ['choice1', 'choice2'], 'projection': 'test_projection', 'nearest': True, 'real_time': True}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            market_data.get_historical(date='2023-08-14', start_date='2022-05-14', end_date='2024-07-17', ticker_list=['test'], ticker_type=['choice1', 'choice2'], projection='test_projection', nearest=True, real_time=True)

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    ###############################################################################
    #  Search Test Case Implementation
    ###############################################################################
    def test_search_is_called(self):
        market_data = MarketData(date=Date(2023, 7, 31))
        expected_data = compress({
            'class_name': 'MarketData',
            'method_name': 'search',
            'self_obj': market_data.to_dict(add_class_path=True),
            'params': {
                'conditions': [['asset', '=', 'test']],
                'fields': ['everysk_id', 'asset'],
                'order_by': '-everysk_id',
                'limit': 10,
                'date': '2023-08-14'
            }
        }, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            market_data.search(
                conditions=[['asset', '=', 'test']],
                fields=['everysk_id', 'asset'],
                order_by='-everysk_id',
                limit=10,
                date='2023-08-14'
            )

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    ###############################################################################
    #  Get Currencies Test Case Implementation
    ###############################################################################
    def test_get_currencies_is_called(self):
        market_data = MarketData()
        expected_data = compress({
            'class_name': 'MarketData',
            'method_name': 'get_currencies',
            'self_obj': market_data.to_dict(add_class_path=True),
            'params': {}
        }, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            market_data.get_currencies()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

#
EVERYSK_SERVER_DEBUG: bool = False

# Used to check access on JSONEndpoints
EVERYSK_SERVER_REST_KEY_NAME: str = 'X-Rest-Key'
EVERYSK_SERVER_REST_KEY_VALUE: str = '12345'

# Used by the GZipMiddleware to compress the response
EVERYSK_SERVER_GZIP_MINIMUM_SIZE: int = 1024
EVERYSK_SERVER_GZIP_COMPRESS_LEVEL: int = 9

# The codes that will be logged by the Logger
EVERYSK_SERVER_CODES_LOG: tuple = (500,)

# URL to use in the redirect response
EVERYSK_SERVER_REDIRECT_URL: str = None



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources import Datastore
from everysk.core.unittests import TestCase, mock


###############################################################################
#   Datastore TestCase Implementation
###############################################################################
class APIDatastoreTestCase(TestCase):

    def test_datastore_class_name(self):
        self.assertEqual(Datastore.class_name(), 'datastore')

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_datastore_explore_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.MagicMock()
        mock_create_api_requestor.return_value = mock_api_requestor
        mock_response = mock.MagicMock()

        mock_api_requestor.post.return_value = mock_response
        mock_kwargs = {
            'api_entry': 'https://example.com',
            'api_version': 'v2',
            'api_sid': 'abc',
            'api_token': 'token_example123',
            'verify_ssl_certs': True
        }
        expected_url = f"/{Datastore.class_name_list()}/explore"
        result = Datastore.explore(**mock_kwargs)

        mock_create_api_requestor.assert_called_once_with(mock_kwargs)
        mock_api_requestor.post.assert_called_once_with(expected_url, mock_kwargs)
        self.assertEqual(result, mock_response)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=unused-import, wildcard-import, unused-wildcard-import

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources.tests import *

from everysk.api._tests.api_requestor import APIRequestorTestCase as EveryskLibAPIRequestorTestCase

from everysk.api._tests.http_client import (
    HTTPClientTestCase as EveryskLibHTTPClientTestCase,
    RequestsClientTestCase as EveryskLibRequestsClientTestCase,
    StandardHTTPClientTestCase as EveryskLibStandardHTTPClientTestCase
)

from everysk.api._tests.init import TestGetAPIConfig as EveryskLibTestGetAPIConfig

from everysk.api._tests.utils import (
    EveryskObjectTestCase as EveryskLibEveryskObjectTestCase,
    EveryskListTestCase as EveryskLibEveryskListTestCase,
    UtilsTestCase as EveryskLibUtilsTestCase

)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Any, Callable, TypedDict
from inspect import isclass

from everysk.config import settings
from everysk.core.exceptions import FieldValueError, InvalidArgumentError
from everysk.core.object import BaseDict, BaseDictConfig
from everysk.core.string import import_from_string

from everysk.sdk.base import BaseSDK

###############################################################################
#   StorageSettings Implementation
###############################################################################
class StorageSettings(TypedDict):
    storage_mode: str
    consistency_check: bool
    skip_validation: bool
    create_fallback: bool

###############################################################################
#   Base Script Implementation
###############################################################################
class Script(BaseDict, BaseSDK):
    """
    A base class for scripted queries.
    This class provides a base implementation for scripted queries.

    Attributes:
        - _klass (callable): The class to instantiate when fetching an entity.

    Example:
        To fetch an entity:
        >>> script = Script(klass=MyEntity)
        >>> entity = script.fetch(user_input, variant, workspace)
    """
    class Config(BaseDictConfig):
        exclude_keys: frozenset[str] = frozenset(['_is_frozen', '_silent', '_errors', '_orderable_attributes'])

    _klass: Callable = None
    _config: Config = None

    def __init__(self, _klass: Callable) -> None:
        super().__init__(_klass=None)

        if _klass is not None and not isclass(_klass):
            try:
                _klass = import_from_string(settings.EVERYSK_SDK_ENTITIES_MODULES_PATH[_klass])
            except KeyError:
                raise FieldValueError(f"The _klass value '{_klass}' must be a class or a string with the class name.") from KeyError

        self._klass = _klass

    def _process__klass(self, value: Any) -> Any:
        """
        This method is used to process the '_klass' attribute.
        """
        return value.__name__

    def _process_entity_output(self, user_input: Any) -> Any:
        """
        Processes the user input and returns an instance of the class or the input itself.

        If the `user_input` is a dictionary or an instance of `BaseDict`, it will be passed as keyword arguments
        to the class constructor (`self._klass`) and an instance of that class will be returned. Otherwise,
        the `user_input` is returned as-is.

        Args:
            user_input (Any): The input data to be processed. This can be a dictionary, an instance of `BaseDict`,
                            or any other type.

        Returns:
            Any: An instance of the class if `user_input` is a dictionary or `BaseDict`; otherwise, the `user_input`
                itself is returned.
        """
        return self._klass(**user_input) if isinstance(user_input, (dict, BaseDict)) else user_input # pylint: disable=not-callable

    def inner_fetch(self, user_input: Any, variant: str, workspace: str = Undefined) -> Any:
        """
        Makes a call to the client to fetch an entity based on user input, variant, and workspace.

        Args:
            user_input (Any): The input provided by the user, which can be used for filtering or as a direct entity ID.
            variant (str): The type of scripted query to execute. Determines how the method processes the user input and constructs the query. Supported variants include 'previousWorkers', 'tagLatest', any string starting with 'select', and potentially others.
            workspace (str): The workspace context for the query. Used for scoping and verifying entity retrieval.

        Returns:
            Any: Depending on the variant and user input, the method might return an entity, or None.

        """
        response = self.get_response(self_obj=self, params={'user_input': user_input, 'variant': variant, 'workspace': workspace})

        return self._process_entity_output(response)

    def fetch(self, user_input: Any, variant: str, workspace: str = Undefined) -> Any:
        """
        Fetches an entity based on user input, variant, and workspace.

        This method provides a way to construct and execute different types of queries
        based on the specified variant. It's designed to handle a variety of scenarios
        and return the desired entity or entities based on the input parameters.

        Args:
            - user_input (Any): The input provided by the user, which can be used for filtering
            or as a direct entity ID.
            - variant (str): The type of scripted query to execute. Determines how the method
            processes the user input and constructs the query. Supported variants include
            'previousWorkers', 'tagLatest', any string starting with 'select', and potentially
            others.
            - workspace (str): The workspace context for the query. Used for scoping and
            verifying entity retrieval.

        Returns:
            - Any: Depending on the variant and user input, the method might return an entity,
            or None.

        Raises:
            - ValueError: If there's an attempted cross-workspace operation or other variant-specific
            error conditions are met.

        Note:
            The method behavior can vary greatly depending on the `variant` parameter, and it's
            important to ensure that the variant aligns with the expected user input structure.

        """
        if user_input is None:
            return None

        if variant == 'previousWorkers' and isinstance(user_input, (dict, BaseDict, self._klass)) and user_input.get('id') is None:
            return self._process_entity_output(user_input)

        entity: Any = self.inner_fetch(user_input, variant, workspace)

        return entity

    def inner_fetch_list(self, user_input: Any, variant: str, workspace: str = Undefined) -> list:
        """
        Fetches a list of entities based on user input, variant, and workspace.

        This method makes a call to the client to perform the necessary query actions.

        Args:
            user_input (Any): The input provided by the user, which can be used for filtering or as a direct entity ID.
            variant (str): The type of scripted query to execute.
            workspace (str): The workspace context for the query.

        Returns:
            list: The list of entities retrieved based on the input parameters.

        """
        return self.get_response(self_obj=self, params={'user_input': user_input, 'variant': variant, 'workspace': workspace})


    def fetch_list(self, user_input: Any, variant: str, workspace: str = Undefined) -> list:
        """
        Fetches an entity based on user input, variant, and workspace.

        This method provides a way to construct and execute different types of queries
        based on the specified variant. It's designed to handle a variety of scenarios
        and return the desired entity or entities based on the input parameters.

        Args:
            user_input (Any): The input provided by the user, which can be used for filtering or as a direct entity ID.
            variant (str): The type of scripted query to execute. Determines how the method processes the user input and constructs the query. Supported variants include 'previousWorkers', 'tagLatest', any string starting with 'select', and potentially others.
            workspace (str): The workspace context for the query. Used for scoping and verifying entity retrieval.

        Returns:
            list: The method will return a list of entities or an empty list.

        Raises:
            ValueError: If there's an attempted cross-workspace operation or other variant-specific error conditions are met.

        Note:
            The method behavior can vary greatly depending on the `variant` parameter, and it's important to ensure that the variant aligns with the expected user input structure.

        """
        if user_input is None:
            return []

        entity_list: list = self.inner_fetch_list(user_input, variant, workspace)

        return [self._klass(**item) for item in entity_list] # pylint: disable=not-callable

    def inner_fetch_multi(self, user_input_list: list[Any], variant_list: list[str], workspace_list: list[str] = Undefined) -> list:
        """
        Fetches a list of entities based on user inputs, query variants, and workspace contexts.

        This method sends a request to the client to execute the necessary query actions,
        dynamically constructed based on the provided parameters.

        Args:
            user_input_list (list[Any]): A list of inputs provided by the user.
            variant_list (list[str]): A list of query variants, each specifying the type of query to execute.
            workspace_list (list[str], optional): A list of workspace contexts, used for scoping and validating
                entity retrieval. Defaults to `Undefined`.

        Returns:
            list: A list of entities retrieved based on the provided parameters. Returns an empty list
            if no entities match the query.
        """
        return self.get_response(self_obj=self, params={'user_input_list': user_input_list, 'variant_list': variant_list, 'workspace_list': workspace_list})


    def fetch_multi(self, user_input_list: list[Any], variant_list: list[str], workspace_list: list[str] = Undefined) -> list:
        """
        Fetches entities based on user inputs, query variants, and workspace contexts.

        This method provides a higher-level interface for constructing and executing different
        types of queries, delegating the core functionality to `inner_fetch_multi`.

        Args:
            user_input_list (list[Any]): A list of inputs provided by the user.
            variant_list (list[str]): A list of query variants, each specifying the type of query to execute.
            workspace_list (list[str], optional): A list of workspace contexts, used for scoping and validating
                entity retrieval. Defaults to `Undefined`.

        Returns:
            list: A list of entities processed and instantiated using the class constructor (`self._klass`).
            Returns an empty list if no entities match the query.
        """
        if not user_input_list:
            return []

        entity_list: list = self.inner_fetch_multi(user_input_list, variant_list, workspace_list)

        return [self._klass(**item) for item in entity_list] # pylint: disable=not-callable

    def persist(self, entity: Any, persist: str, consistency_check: bool = False) -> Any:
        """
        This method provides a way to persist an entity based on the specified persist type.

        Args:
            - entity (Any): The entity to persist.
            - persist (str): The type of persist to execute. Determines how the method
            persists the entity. Supported persists include 'insert', 'update', and 'delete'.
            - consistency_check (bool): A flag to enable consistency checks before persisting.

        Returns:
            - Any: Depending on the persist type, the method might return an entity.

        """
        response = self.get_response(self_obj=self, params={'entity': entity, 'persist': persist, 'consistency_check': consistency_check})

        if isinstance(response, dict):
            response = self._klass(**response) # pylint: disable=not-callable

        return response

    def inner_storage(self, entity: Any, storage_settings: StorageSettings) -> Any:
        """
        Handles the storage actions for a given entity using specified storage settings.

        This method makes a call to the client to perform the necessary storage actions.

        Args:
            entity (Any): The entity to be stored.
            storage_settings (StorageSettings): The settings and configurations for storage.

        Returns:
            Any: The response from the client after performing the storage actions.

        """
        response = self.get_response(self_obj=self, params={'entity': entity, 'storage_settings': storage_settings})

        return self._process_entity_output(response)


    def storage(self, entity: Any, storage_settings: StorageSettings) -> Any:
        """
        Stores an entity based on the specified storage settings.

        This method determines the appropriate storage actions for an entity, including handling
        different storage modes, performing consistency checks, and optionally skipping validation.

        Args:
            entity (Any): The entity to be stored.
            storage_settings (StorageSettings): A configuration object containing:
                - storage_mode (str): The storage mode to use. Supported modes include 'transient', 'create', and 'update'.
                - consistency_check (bool): A flag indicating whether to perform a consistency check.
                - validate (bool): A flag indicating whether to validate the entity parameters or not.
                - create_fallback (bool): A flag indicating whether to create an entity if it does not exist.

        Returns:
            Any: Depending on the storage mode, the method might return the stored entity or a list of entities.

        Raises:
            InvalidArgumentError: If the entity is empty.

        """
        if not entity:
            raise InvalidArgumentError('Entity should not be empty.')

        if storage_settings['storage_mode'] == 'transient' and not storage_settings['validate']:
            return self._process_entity_output(entity)

        entity: Any = self.inner_storage(entity, storage_settings)

        return entity



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Self, Any

from everysk.config import settings
from everysk.core.fields import StrField, ListField, BoolField, DictField, IntField, DateTimeField, FloatField
from everysk.core.object import BaseDict
from everysk.core.datetime import DateTime

from everysk.sdk.entities.base import BaseEntity
from everysk.sdk.entities.query import Query

###############################################################################
#   Result Implementation
###############################################################################
class Result(BaseDict):
    status = StrField(default='ERROR', choices=('OK', 'ERROR', 'UNKNOW'))
    data: list | dict | None = None
    log = ListField()

class ResultField(DictField):
    attr_type: Result | BaseDict | dict = Result | BaseDict | dict

    def clean_value(self, value: Any) -> Result | None:
        """
        This method cleans the value of the field.

        Args:s
            value (Any): The value to be cleaned.

        Returns:
            Securities: The cleaned value.
        """
        if isinstance(value, dict):
            value = Result(**value)

        return super().clean_value(value)

###############################################################################
#   ParallelInfo Implementation
###############################################################################
class ParallelInfo(BaseDict):
    index = IntField(default=settings.WORKER_EXECUTION_UNFORKED_PARALLEL_INDEX, required_lazy=True)
    length = IntField(default=settings.WORKER_EXECUTION_UNFORKED_PARALLEL_LENGTH, required_lazy=True)

class ParallelInfoField(DictField):
    attr_type: ParallelInfo | BaseDict | dict = ParallelInfo | BaseDict | dict

    def __init__(self, default: Any = None, required: bool = False, readonly: bool = False, required_lazy: bool = True, empty_is_none: bool = False, **kwargs) -> None:
        if default is None:
            default = ParallelInfo()
        super().__init__(default, required, readonly, required_lazy, empty_is_none, **kwargs)

    def clean_value(self, value: Any) -> Result | None:
        """
        This method cleans the value of the field.

        Args:s
            value (Any): The value to be cleaned.

        Returns:
            Securities: The cleaned value.
        """
        if isinstance(value, (dict, BaseDict)):
            value = ParallelInfo(**value)

        return super().clean_value(value)

###############################################################################
#   InputParams Implementation
###############################################################################
class InputParams(BaseDict):
    worker_id = StrField(empty_is_none=True)
    workflow_id = StrField(regex=settings.WORKFLOW_ID_REGEX, empty_is_none=True)

    worker_execution_id = StrField(regex=settings.WORKER_EXECUTION_ID_REGEX, empty_is_none=True)
    workflow_execution_id = StrField(regex=settings.WORKFLOW_EXECUTION_ID_REGEX, empty_is_none=True)

    workspace = StrField(empty_is_none=True)
    worker_type = StrField(default=Undefined, choices=settings.WORKER_TEMPLATE_TYPES)
    script_inputs = DictField()
    inputs_info = DictField()
    parallel_info = ParallelInfoField()

    def to_dict(self, add_class_path: bool = False, recursion: bool = False) -> dict:
        """
        This method is used to convert the object to a dictionary.
        """
        dct: dict = super().to_dict(add_class_path=add_class_path, recursion=recursion)

        if isinstance(dct['parallel_info'], ParallelInfo):
            dct['parallel_info'] = dct['parallel_info'].to_dict(add_class_path=add_class_path, recursion=recursion)

        return dct

class InputParamsField(DictField):
    attr_type: InputParams | BaseDict | dict = InputParams | BaseDict | dict

    def clean_value(self, value: Any) -> InputParams | None:
        """
        This method cleans the value of the field.

        Args:s
            value (Any): The value to be cleaned.

        Returns:
            Securities: The cleaned value.
        """
        if isinstance(value, dict):
            value = InputParams(**value)

        return super().clean_value(value)

###############################################################################
#   Worker Execution Implementation
###############################################################################
class WorkerExecution(BaseEntity):

    id = StrField(regex=settings.WORKER_EXECUTION_ID_REGEX, required_lazy=True, empty_is_none=True)

    result = ResultField()
    storage = BoolField(default=False)
    input_params = InputParamsField(required_lazy=True)
    storage_input_params = BoolField(default=False)

    parallel_info = ParallelInfoField()
    status = StrField(default=settings.WORKER_EXECUTION_STATUS_PREPARING, choices=settings.WORKER_EXECUTION_STATUS_LIST)
    execution_type = StrField(default=Undefined, choices=settings.WORKER_EXECUTION_TYPE_LIST, required_lazy=True)
    start_time = DateTimeField()
    end_time = DateTimeField()
    duration = FloatField(default=0.0)
    cpu_time = FloatField(default=0.0)
    process_cpu_time = FloatField(default=0.0)

    workflow_execution_id = StrField(regex=settings.WORKFLOW_EXECUTION_ID_REGEX, required_lazy=True, empty_is_none=True)
    workflow_id = StrField(regex=settings.WORKFLOW_ID_REGEX, required_lazy=True, empty_is_none=True)
    workflow_name = StrField(required_lazy=True, empty_is_none=True)
    worker_id = StrField(regex=settings.WORKER_ID_REGEX, required_lazy=True, empty_is_none=True)
    worker_name = StrField(required_lazy=True, empty_is_none=True)
    worker_type = StrField(default=Undefined, choices=settings.WORKER_TEMPLATE_TYPES)

    @staticmethod
    def get_id_prefix() -> str:
        """
        Returns the prefix of the Worker Execution id field value.

        Returns:
            str: The prefix of the Worker Execution id field value.

        Usage:
            >>> WorkerExecution.get_id_prefix()
            'wkex_'

        Notes:
            The prefix is typically used to distinguish Worker Execution IDs from other types of IDs.
        """
        return settings.WORKER_EXECUTION_ID_PREFIX

    def generate_id(self) -> str:
        """
        Generate a unique ID for an entity instance.

        Returns:
            str: The generated unique ID.

        Example:
            To generate a unique ID for an entity instance:
            >>> unique_id = MyEntity().generate_id()
        """
        raise NotImplementedError()

    def to_dict(self, add_class_path: bool = False, recursion: bool = False) -> dict:
        """
        This method is used to convert the object to a dictionary.
        """
        dct: dict = super().to_dict(add_class_path=add_class_path, recursion=recursion)

        if isinstance(dct['parallel_info'], ParallelInfo):
            dct['parallel_info'] = dct['parallel_info'].to_dict(add_class_path=add_class_path, recursion=recursion)

        if isinstance(dct['result'], Result):
            dct['result'] = dct['result'].to_dict(add_class_path=add_class_path, recursion=recursion)

        if isinstance(dct['input_params'], InputParams):
            dct['input_params'] = dct['input_params'].to_dict(add_class_path=add_class_path, recursion=recursion)

        if add_class_path is False:

            if self.duration is None:
                now = DateTime.now()
                start = self.start_time if self.start_time else now
                end = self.end_time if self.end_time else now
                dct['duration'] = (end - start).total_seconds()

            dct['started'] = self.start_time.timestamp() if self.start_time else None
            dct['trigger'] = self.execution_type

            if dct['result'] is not None:
                dct['result'] = dct.get('result', {}).get('data', None)

            dct.pop('execution_type')
            dct.pop('start_time')
            dct.pop('end_time')
            dct.pop('storage')
            dct.pop('storage_input_params')

        return dct

    def _check_entity_to_query(self) -> bool:
        """
        Check the entity object to query.

        Returns:
            bool: True if the entity object is valid.
        """
        return True

    def _check_query(self, query: Query) -> bool:
        """
        Check the query object.

        Args:
            query (Query): The query object.

        Returns:
            bool: True if the query object is valid.
        """
        return True

    def _mount_query(self, query: Query) -> Query:
        """
        Mount the query object.

        Args:
            query (Query): The query object.

        Returns:
            Query: The query object.
        """
        if self.workflow_execution_id is not None:
            query = query.where('workflow_execution_id', self.workflow_execution_id)

        return query

    @classmethod
    def get_input_params(cls, entity_id: str) -> InputParams:
        """
        Get the input_params of a Worker Execution with the given ID.

        Args:
            entity_id (str): The ID of the Worker Execution.

        Returns:
            InputParams: The input_params of the Worker Execution.

        Raises:
            ValueError: If the entity ID is invalid.
            ValueError: If the entity is not found.
        """
        if cls.validate_id(entity_id) is False:
            raise ValueError(f'Invalid Entity ID: {entity_id}')

        entity: Self | None = cls.retrieve(entity_id)

        if entity is None:
            raise ValueError(f'Entity not found. Entity ID: {entity_id}')

        if entity.input_params.worker_execution_id is None:
            entity.input_params.worker_execution_id = entity.id # pylint: disable=attribute-defined-outside-init

        return entity.input_params



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['create_application']
from starlette.applications import Starlette

from everysk.config import settings
from everysk.server.middlewares import Middleware, update_with_default_middlewares
from everysk.server.routing import Route


def create_application(
        *, # This disables positional arguments for the function
        routes: list[Route],
        debug: bool = settings.EVERYSK_SERVER_DEBUG,
        middlewares: list[Middleware] = None,
        exception_handlers: dict[int | Exception, callable] = None
    ) -> Starlette:
    """
    Create a Starlette application with the given routes, middleware and exception handlers.
    By default the first middleware is the GZipMiddleware and SecurityHeadersMiddleware is the last one,
    so you don't need to add them in the list.
    Routes could be normal Starlette routes or the RouteLazy class that will lazy load the endpoint class.

    Args:
        routes (list[Route]): A list of routes to serve incoming HTTP and WebSocket requests.
        debug (bool, optional): Boolean indicating if debug tracebacks should be returned on errors. Defaults to False.
        middleware (list[Middleware], optional): A list of middleware to run for every request. Defaults to None.
        exception_handlers (dict[int, callable], optional): A mapping of either integer status codes, or exception class types onto callables which handle the exceptions. Defaults to None.

    """
    middlewares = update_with_default_middlewares(middlewares)

    return Starlette(debug=debug, routes=routes, middleware=middlewares, exception_handlers=exception_handlers)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import DateTime, ZoneInfo
from everysk.core.exceptions import SDKValueError
from everysk.core.http import HttpSDKPOSTConnection, httpx
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.report.base import Report


###############################################################################
#   Report TestCase Implementation
###############################################################################
class ReportTestCase(TestCase):

    def setUp(self):
        self.sample_data = {
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'version': 'v1',
            'id': 'repo_12345678',
            'name': 'My Report',
            'description': 'This is a sample report.',
            'tags': ['tag1',
            'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': DateTime(2022, 1, 1, 12, 0, tzinfo=ZoneInfo(key='UTC')),
            'widgets': [{'type': 'chart', 'data': {}}],
            'url': '/sefdsf5s54sdfsksdfs5',
            'authorization': 'private',
            'config_cascaded': {'setting1': 'value1',
            'setting2': 'value2'},
            'layout_content': {'section1': {}}
        }
        self.report = Report(**self.sample_data)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_get_id_prefix(self):
        self.assertEqual(Report.get_id_prefix(), settings.REPORT_ID_PREFIX)

    def test_validate(self):
        expected_data = compress({'class_name': 'Report', 'method_name': 'validate', 'self_obj': self.report.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')
        report: Report = self.report.copy()

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            report.validate()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_query_load_with_id(self):
        expected_data = compress({'class_name': 'Report', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'repo_1234567891011211234567890', 'projection': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Report(id='repo_1234567891011211234567890', workspace='SampleWorkspace').load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_query_load(self):
        report = Report(link_uid='SampleLinkUID', workspace='SampleWorkspace')
        expected_data = compress({'class_name': 'Query', 'method_name': 'load', 'self_obj': report.to_query().to_dict(add_class_path=True), 'params': {'offset': None}}, protocol='gzip', serialize='json')

        with mock.patch('httpx.Client.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            report.load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            data=expected_data
        )

    def test_entity_to_query(self):
        cmp_query = Report.query.where('url', self.report.url)
        query = Report(url=self.report.url).to_query() #pylint: disable=protected-access

        self.assertDictEqual(query, cmp_query)

    def test_check_query_raises_sdk_error(self):
        with self.assertRaises(SDKValueError) as context:
            Report(url=self.report.url, name='report_name').to_query()

        self.assertEqual(str(context.exception), "Can't filter by URL and Name, Tags or Link UID at the same time")

    def test_to_dict_report(self):
        result = self.report.to_dict()
        expected_data = {
            'version': 'v1',
            'id': 'repo_12345678',
            'name': 'My Report',
            'description': 'This is a sample report.',
            'tags': ['tag1',
            'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': '20220101',
            'widgets': [{'type': 'chart',
            'data': {}}],
            'url': 'https://app.everysk.com/report/sefdsf5s54sdfsksdfs5',
            'authorization': 'PRIVATE',
            'date_time': '20220101 12:00:00',
            'created': 1694250549,
            'updated': 1694250549,
            'absolute_url': 'https://app.everysk.com/report/sefdsf5s54sdfsksdfs5',
            'relative_url': '/report/sefdsf5s54sdfsksdfs5'
        }

        self.assertIsInstance(result, dict)
        self.assertEqual(result, expected_data)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=unused-import

## Remember to prefix all import with EveryskLib to avoid clash with other tests

## Cloud function Test Cases
try:
    from everysk.core.cloud_function.tests import CloudFunctionTestCase as EveryskLibCloudFunctionTestCase
except ModuleNotFoundError as error:
    # This will prevent running these tests if redis is not installed
    if not error.args[0].startswith("No module named 'redis'"):
        raise error

## Compress Test Cases
from everysk.core._tests.compress import (
    CompressTestCase as EveryskLibCompressTestCase,
    CompressGzipJsonTestCase as EveryskLibCompressGzipJsonTestCase,
    CompressGzipPickleTestCase as EveryskLibCompressGzipPickleTestCase,
    CompressZlibJsonTestCase as EveryskLibCompressZlibJsonTestCase,
    CompressZlibPickleTestCase as EveryskLibCompressZlibPickleTestCase,
    FileHandlingTestCase as EveryskLibFileHandlingTestCase
)

## Config Test Cases
from everysk.core._tests.config import (
    SettingsModulesTestCase as EveryskLibSettingsModulesTestCase,
    SettingsTestCase as EveryskLibSettingsTestCase,
    SettingsManagerTestCase as EveryskLibSettingsManagerTestCase
)

## Date, DateTime Test Cases
from everysk.core.datetime.tests.date import DateTestCase as EveryskLibDateTestCase
from everysk.core.datetime.tests.datetime import DateTimeTestCase as EveryskLibDateTimeTestCase
from everysk.core.datetime.tests.date_mixin import GetHolidaysTestCase as EveryskLibDateMixinGetHolidaysTestCase
from everysk.core.datetime.tests.calendar import CalendarTestCase as EveryskLibCalendarTestCase

## Exceptions Test Cases
from everysk.core._tests.exceptions import (
    BaseExceptionTestCase as EveryskLibBaseExceptionTestCase, DefaultErrorTestCase as EveryskLibDefaultErrorTestCase,
    HttpErrorTestCase as EveryskLibHttpErrorTestCase, FieldValueErrorTestCase as EveryskLibFieldValueErrorTestCase,
    ReadonlyErrorTestCase as EveryskLibReadonlyErrorTestCase, RequiredErrorTestCase as EveryskLibRequiredErrorTestCase,
    TestAPIError as EveryskLibTestAPIError,
    HandledExceptionTestCase as EveryskLibHandledExceptionTestCase,
    SDKExceptionsTestCase as EveryskLibSDKExceptionsTestCase
)

## Fields Test Cases
from everysk.core._tests.fields import (
    BoolFieldTestCase as EveryskLibBoolFieldTestCase, ChoiceFieldTestCase as EveryskLibChoiceFieldTestCase, DateFieldTestCase as EveryskLibDateFieldTestCase,
    DateTimeFieldTestCase as EveryskLibDateTimeFieldTestCase, DictFieldTestCase as EveryskLibDictFieldTestCase, FieldTestCase as EveryskLibFieldTestCase,
    FieldUndefinedTestCase as EveryskLibFieldUndefinedTestCase, FloatFieldTestCase as EveryskLibFloatFieldTestCase, IntFieldTestCase as EveryskLibIntFieldTestCase,
    IteratorFieldTestCase as EveryskLibIteratorFieldTestCase, ListFieldTestCase as EveryskLibListFieldTestCase, StrFieldTestCase as EveryskLibStrFieldTestCase,
    TupleFieldTestCase as EveryskLibTupleFieldTestCase, ObjectInitPropertyTestCase as EveryskLibObjectInitPropertyTestCase, COD3770TestCase as EveryskLibCOD3770TestCase,
    URLFieldTestCase as EveryskLibURLFieldTestCase, SetFieldTestCase as EveryskLibSetFieldTestCase, EmailFieldTestCase as EveryskLibEmailFieldTestCase
)

## Firestore Test Cases
try:
    from everysk.core._tests.firestore import (
        BaseDocumentCachedConfigTestCase as EveryskLibBaseDocumentCachedConfigTestCase,
        BaseDocumentConfigTestCase as EveryskLibBaseDocumentConfigTestCase,
        DocumentCachedTestCase as EveryskLibDocumentCachedTestCase,
        DocumentTestCase as EveryskLibDocumentTestCase,
        FirestoreClientTestCase as EveryskLibFirestoreClientTestCase,
        LoadsPaginatedTestCase as EveryskLibLoadsPaginatedTestCase
    )
except ModuleNotFoundError as error:
    # This will prevent running these tests if google-cloud-firestore is not installed
    if not error.args[0].startswith("No module named 'google"):
        raise error

## Http Test Cases
try:
    from everysk.core._tests.http import (
        HttpConnectionTestCase as EveryskLibHttpConnectionTestCase,
        HttpConnectionConfigTestCase as EveryskLibHttpConnectionConfigTestCase,
        HttpGETConnectionTestCase as EveryskLibHttpGETConnectionTestCase,
        HttpPOSTConnectionTestCase as EveryskLibHttpPOSTConnectionTestCase,
        HttpSDKPOSTConnectionTestCase as EveryskLibHttpSDKPOSTConnectionTestCase,
        HttpPOSTCompressedConnectionTestCase as EveryskLibHttpPOSTCompressedConnectionTestCase,
        HttpDELETEConnectionTestCase as EveryskLibHttpDELETEConnectioNTestCase,
        HttpHEADConnectionTestCase as EveryskLibHttpHEADConnectionTestCase,
        HttpOPTIONSConnectionTestCase as EveryskLibHttpOPTIONSConnectionTestCase,
        HttpPATCHConnectionTestCase as EveryskLibHttpPATCHConnectionTestCase,
        HttpPUTConnectionTestCase as EveryskLibHttpPUTCompressedConnectionTestCase
    )
except ModuleNotFoundError as error:
    # This will prevent running these tests if requests is not installed
    if not error.args[0].startswith("No module named 'requests'"):
        raise error

## Log Test Cases
from everysk.core._tests.log import (
    LoggerExtraDataTestCase as EveryskLibLoggerExtraDataTestCase,
    LoggerFlaskTestCase as EveryskLibLogFlaskTestCase,
    LoggerFormatterTestCase as EveryskLibLoggerFormatterTestCase,
    LoggerManagerTestCase as EveryskLibLoggerManagerTestCase,
    LoggerMethodsTestCase as EveryskLibLoggerMethodsTestCase,
    LoggerTestCase as EveryskLibLoggerTestCase,
    LoggerTraceTestCase as EveryskLibLogTraceTestCase,
    LoggerStdoutTestCase as EveryskLibLoggerStdoutTestCase,
    LoggerJsonTestCase as EveryskLibLoggerJsonTestCase,
)
try:
    from everysk.core._tests.log import LoggerSlackTestCase as EveryskLibLoggerSlackTestCase # We need requests to run this test
except ModuleNotFoundError as error:
    # This will prevent running these tests if requests is not installed
    if not error.args[0].startswith("No module named 'requests'"):
        raise error


## Object Test Cases
from everysk.core._tests.object import (
    BaseDictPropertyTestCase as EveryskLibBaseDictPropertyTestCase, BaseDictTestCase as EveryskLibBaseDictTestCase,
    BaseFieldTestCase as EveryskLibBaseFieldTestCase, BaseObjectTestCase as EveryskLibBaseObjectTestCase,
    ConfigHashTestCase as EveryskLibConfigHashTestCase, FrozenDictTestCase as EveryskLibFrozenDictTestCase,
    FrozenObjectTestCase as EveryskLibFrozenObjectTestCase, MetaClassConfigTestCase as EveryskLibMetaClassConfigTestCase,
    RequiredTestCase as EveryskLibRequiredTestCase, ValidateTestCase as EveryskLibValidateTestCase,
    MetaClassAttributesTestCase as EveryskLibMetaClassAttributesTestCase,
    NpArrayTestCase as EveryskLibNpArrayTestCase, AfterInitTestCase as EveryskLibAfterInitTestCase,
    BeforeInitTestCase as EveryskLibBeforeInitTestCase, SilentTestCase as EveryskLibSilentTestCase,
    TypingCheckingTestCase as EveryskLibTypingCheckingTestCase, BaseDictSuperTestCase as EveryskLibBaseDictSuperTestCase
)

## Number Test Cases
from everysk.core._tests.number import NumberTestCase as EveryskLibNumberTestCase

## Signing Test Cases
from everysk.core._tests.signing import SignTestCase as EveryskLibSignTestCase, UnsignTestCase as EveryskLibUnsignTestCase

## String Test Cases
from everysk.core._tests.string import StringTestCase as EveryskLibStringTestCase

## Redis Test Cases
try:
    from everysk.core._tests.redis import (
        RedisCacheCompressedTestCase as EveryskLibRedisCacheCompressedTestCase,
        RedisCacheTestCase as EveryskLibRedisCacheTestCase, RedisChannelTestCase as EveryskLibRedisChannelTestCase,
        RedisClientTestCase as EveryskLibRedisClientTestCase, RedisListTestCase as EveryskLibRedisListTestCase,
        RedisLockTestCase as EveryskLibRedisLockTestCase, RedisCacheGetSetTestCase as EveryskLibRedisCacheGetSetTestCase,
        CacheDecoratorTestCase as EveryskLibCacheDecoratorTestCase
    )
except ModuleNotFoundError as error:
    # This will prevent running these tests if redis is not installed
    if not error.args[0].startswith("No module named 'redis'"):
        raise error

## Serialize Test Cases
from everysk.core._tests.serialize import SerializeDumpsTestCase as EveryskLibSerializeDumpsTestCase, SerializeLoadsTestCase as EveryskLibSerializeLoadsTestCase

## SFTP Test Cases
try:
    from everysk.core._tests.sftp import KnownHostsTestCase as EveryskLibKnownHostsTestCase, SFTPTestCase as EveryskLibSFTPTestCase
except ModuleNotFoundError as error:
    # This will prevent running these tests if Paramiko is not installed
    if not error.args[0].startswith("No module named 'paramiko'"):
        raise error

## Slack Test Cases
try:
    from everysk.core._tests.slack import SlackTestCase as EveryskLibSlackTestCase
except ModuleNotFoundError as error:
    # This will prevent running these tests if requests is not installed
    if not error.args[0].startswith("No module named 'requests'"):
        raise error


## Thread Test Cases
from everysk.core._tests.threads import ThreadPoolTestCase as EveryskLibThreadPoolTestCase, ThreadTestCase as EveryskLibThreadTestCase

## Undefined Test Cases
from everysk.core._tests.undefined import UndefinedTestCase as EveryskLibUndefinedTestCase

## Unittest Test Cases
from everysk.core._tests.unittests import SDKUnittestTestCase as EveryskLibSDKUnittestTestCase

## Utils Test Cases
from everysk.core._tests.utils import (
    BoolConverterTestCase as EveryskLibBoolConverterTestCase,
    SearchKeyTestCase as EveryskLibSearchKeyTestCase
)

## Lists Test Cases
from everysk.core._tests.lists import (
    SplitInSlicesTestCase as EveryskLibSplitInSlicesTestCase,
    SlicesTestCase as EveryskLibSlicesTestCase,
    SortListDictTestCase as EveryskLibSortListDictTestCase
)

## Workers Test Cases
try:
    from everysk.core._tests.workers import (
        BaseGoogleTestCase as EveryskLibBaseGoogleTestCase,
        TaskGoogleTestCase as EveryskLibTaskGoogleTestCase,
        WorkerGoogleTestCase as EveryskLibWorkerGoogleTestCase
    )
except ModuleNotFoundError as error:
    # This will prevent running these tests if google-cloud-tasks is not installed
    if not error.args[0].startswith("No module named 'google"):
        raise error



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.fields import StrField, DateTimeField, ListField, FloatField
from everysk.sdk.entities.base import BaseEntity
from everysk.sdk.entities.query import Query

###############################################################################
#   Worker Execution Implementation
###############################################################################
class WorkflowExecution(BaseEntity):

    id = StrField(regex=settings.WORKFLOW_EXECUTION_ID_REGEX, required_lazy=True, empty_is_none=True)

    # This is a legacy status used in the old version of the API library.
    status = StrField(default=settings.WORKFLOW_EXECUTION_STATUS_OK, choices=settings.WORKFLOW_EXECUTION_STATUS, required_lazy=True)
    run_status = StrField(default=settings.WORKFLOW_EXECUTION_STATUS_PREPARING, choices=settings.WORKFLOW_EXECUTION_STATUS, required_lazy=True)

    execution_type = StrField(default=Undefined, choices=settings.WORKER_EXECUTION_TYPE_LIST, required_lazy=True)
    start_time = DateTimeField()
    end_time = DateTimeField()
    duration = FloatField()
    real_execution_time = FloatField()
    total_execution_time = FloatField()

    workflow_id = StrField(regex=settings.WORKFLOW_ID_REGEX, required_lazy=True, empty_is_none=True)
    workflow_name = StrField(required_lazy=True)
    workspace = StrField(required_lazy=False, empty_is_none=True)

    started_worker_id = StrField(regex=settings.WORKER_ID_REGEX, required_lazy=True, empty_is_none=True)
    ender_worker_id = StrField(regex=settings.WORKER_ID_REGEX, empty_is_none=True)
    ender_worker_execution_id = StrField(regex=settings.WORKER_EXECUTION_ID_REGEX, empty_is_none=True)

    worker_ids = ListField()
    resume = ListField()

    @staticmethod
    def get_id_prefix() -> str:
        """
        Returns the prefix of the Worker Execution id field value.

        Returns:
            str: The prefix of the Worker Execution id field value.

        Usage:
            >>> WorkflowExecution.get_id_prefix()
            'wfex_'

        Notes:
            The prefix is typically used to distinguish Worker Execution IDs from other types of IDs.
        """
        return settings.WORKFLOW_EXECUTION_ID_PREFIX

    def to_dict(self, add_class_path: bool = False, recursion: bool = False) -> dict:
        """
        This method is used to convert the object to a dictionary.
        """
        dct: dict = super().to_dict(add_class_path=add_class_path, recursion=recursion)

        if add_class_path is False:
            dct['started'] = self.start_time.timestamp() if self.start_time else None
            dct['trigger'] = self.execution_type

            dct.pop('worker_ids')
            dct.pop('execution_type')
            dct.pop('start_time')
            dct.pop('end_time')

        return dct

    def _check_entity_to_query(self) -> bool:
        """
        Check the entity object to query.

        Returns:
            bool: True if the entity object is valid.
        """
        return True

    def _check_query(self, query: Query) -> bool:
        """
        Check the query object.

        Args:
            query (Query): The query object.

        Returns:
            bool: True if the query object is valid.
        """
        return True

    def _mount_query(self, query: Query) -> Query:
        """
        Mount the query object.

        Args:
            query (Query): The query object.

        Returns:
            Query: The query object.
        """
        if self.workflow_id is not None:
            query = query.where('workflow_id', self.workflow_id)

        return query
